""" Classes for representing hierarchical language structures, such as syntax
trees and morphological trees. This is an adaptation of the original tree.py
file from NLTK. Probabilistic trees have been removed, as well as the
possibility to read off CFG productions. """
# Original notice:
# Natural Language Toolkit: Text Trees
#
# Copyright (C) 2001-2010 NLTK Project
# Author: Edward Loper <edloper@gradient.cis.upenn.edu>
#         Steven Bird <sb@csse.unimelb.edu.au>
#         Nathan Bodenstab <bodenstab@cslu.ogi.edu> (tree transforms)
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

from __future__ import division, print_function, unicode_literals
import re
import sys
if sys.version[0] >= '3':
	basestring = str  # pylint: disable=W0622,C0103


class Tree(list):
	""" A hierarchical structure. Each Tree represents a single hierarchical
	grouping of leaves and subtrees. For example, each constituent in a syntax
	tree is represented by a single Tree.
	A tree's children are encoded as a list of leaves and subtrees, where a
	leaf is a basic (non-tree) value; and a subtree is a nested Tree.
	Any other properties that a Tree defines are known as node properties, and
	are used to add information about individual hierarchical groupings. For
	example, syntax trees use a label property to label syntactic constituents
	with phrase tags, such as "NP" and "VP".
	Several Tree methods use tree positions to specify children or descendants
	of a tree. Tree positions are defined as follows:

	- The tree position ``i`` specifies a Tree's ith child.
	- The tree position () specifies the Tree itself.
	- If ``p`` is the tree position of descendant d, then
		``p + (i,)`` specifies the ith child of d.

	i.e., every tree position is either a single index ``i``,
	specifying ``self[i]``; or a sequence ``(i1, i2, ..., iN)``,
	specifying ``self[i1][i2]...[iN]``.

	The constructor can be called in two ways:

	- ``Tree(label, children)`` constructs a new tree with the specified label
		and list of children.
	- ``Tree(s)`` constructs a new tree by parsing the string s. Equivalent to
		calling the class method ``Tree.parse(s)``. """
	def __new__(cls, label_or_str=None, children=None):
		if label_or_str is None:
			return list.__new__(cls)  # used by copy.deepcopy
		if children is None:
			if not isinstance(label_or_str, basestring):
				raise TypeError("%s: Expected a label and child list "
						"or a single string" % cls.__name__)
			return cls.parse(label_or_str)
		if (isinstance(children, basestring) or
				not hasattr(children, '__iter__')):
			raise TypeError("%s() argument 2 should be a list, not a "
					"string" % cls.__name__)
		return list.__new__(cls, label_or_str, children)

	def __init__(self, label_or_str, children=None):
		# Because __new__ may delegate to Tree.parse(), the __init__
		# method may end up getting called more than once (once when
		# constructing the return value for Tree.parse; and again when
		# __new__ returns). We therefore check if `children` is None
		# (which will cause __new__ to call Tree.parse()); if so, then
		# __init__ has already been called once, so just return.
		if children is None:
			return
		list.__init__(self, children)
		self.label = label_or_str
		self.source = self.bitset = None

	# === Comparison operators ==================================
	def __eq__(self, other):
		if not isinstance(other, Tree):
			return False
		return self.label == other.label and list.__eq__(self, other)

	def __ne__(self, other):
		return not (self == other)

	def __lt__(self, other):
		if not isinstance(other, Tree):
			return False
		return self.label < other.label or list.__lt__(self, other)

	def __le__(self, other):
		if not isinstance(other, Tree):
			return False
		return self.label <= other.label or list.__le__(self, other)

	def __gt__(self, other):
		if not isinstance(other, Tree):
			return True
		return self.label > other.label or list.__gt__(self, other)

	def __ge__(self, other):
		if not isinstance(other, Tree):
			return False
		return self.label >= other.label or list.__ge__(self, other)

	# === Disabled list operations ==============================
	def __mul__(self, _):
		raise TypeError('Tree does not support multiplication')

	def __rmul__(self, _):
		raise TypeError('Tree does not support multiplication')

	def __add__(self, _):
		raise TypeError('Tree does not support addition')

	def __radd__(self, _):
		raise TypeError('Tree does not support addition')

	# === Indexing (with support for tree positions) ============
	def __getitem__(self, index):
		if isinstance(index, (int, slice)):
			return list.__getitem__(self, index)
		else:
			if len(index) == 0:
				return self
			elif len(index) == 1:
				return self[int(index[0])]
			return self[int(index[0])][index[1:]]

	def __setitem__(self, index, value):
		if isinstance(index, (int, slice)):
			return list.__setitem__(self, index, value)
		else:
			if len(index) == 0:
				raise IndexError('The tree position () may not be '
						'assigned to.')
			elif len(index) == 1:
				self[index[0]] = value
			else:
				self[index[0]][index[1:]] = value

	def __delitem__(self, index):
		if isinstance(index, (int, slice)):
			return list.__delitem__(self, index)
		else:
			if len(index) == 0:
				raise IndexError('The tree position () may not be deleted.')
			elif len(index) == 1:
				del self[index[0]]
			else:
				del self[index[0]][index[1:]]

	# === Basic tree operations =================================
	def leaves(self):
		""" :returns: list containing this tree's leaves. The order reflects \
		the order of the leaves in the tree's hierarchical structure. """
		leaves = []
		for child in self:
			if isinstance(child, Tree):
				leaves.extend(child.leaves())
			else:
				leaves.append(child)
		return leaves

	def flatten(self):
		""" :returns: a tree consisting of this tree's root connected \
		directly to its leaves, omitting all internal nodes. """
		return Tree(self.label, self.leaves())

	def height(self):
		"""	:returns: The height of this tree. The height of a tree \
		containing no children is 1; the height of a tree containing only \
		leaves is 2; and the height of any other tree is one plus the maximum \
		of its children's heights. """
		max_child_height = 0
		for child in self:
			if isinstance(child, Tree):
				max_child_height = max(max_child_height, child.height())
			else:
				max_child_height = max(max_child_height, 1)
		return 1 + max_child_height

	def depth(self):
		""" The depth of a tree is its height - 1. """
		return self.height() - 1

	def treepositions(self, order='preorder'):
		""" :param order: One of preorder, postorder, bothorder, leaves. """
		positions = []
		if order in ('preorder', 'bothorder'):
			positions.append(())
		for i, child in enumerate(self):
			if isinstance(child, Tree):
				childpos = child.treepositions(order)
				positions.extend((i, ) + p for p in childpos)
			else:
				positions.append((i, ))
		if order in ('postorder', 'bothorder'):
			positions.append(())
		return positions

	def subtrees(self, condition=None):
		""" Generate all the subtrees of this tree, optionally restricted
		to trees matching the condition function. NB: if the condition does not
		match a subtree, its children are still considered.

		:param condition: the function to filter all local trees """
		if condition is None or condition(self):
			yield self
		for child in self:
			if isinstance(child, Tree):
				for subtree in child.subtrees(condition):
					yield subtree

	def pos(self):
		""" :returns: a list of tuples containing leaves and pre-terminals \
		(part-of-speech tags). The order reflects the order of the leaves in \
		the tree's hierarchical structure. """
		pos = []
		for child in self:
			if isinstance(child, Tree):
				pos.extend(child.pos())
			else:
				pos.append((child, self.label))
		return pos

	def leaf_treeposition(self, index):
		""" :returns: The tree position of the index-th leaf in this tree; \
		i.e., if tp=self.leaf_treeposition(i), then self[tp]==self.leaves()[i].
		:raises IndexError: if this tree contains fewer than index+1 leaves, \
		or if index<0. """
		if index < 0:
			raise IndexError('index must be non-negative')
		stack = [(self, ())]
		while stack:
			value, treepos = stack.pop()
			if not isinstance(value, Tree):
				if index == 0:
					return treepos
				else:
					index -= 1
			else:
				for i in range(len(value) - 1, -1, -1):
					stack.append((value[i], treepos + (i, )))
		raise IndexError('index must be less than or equal to len(self)')

	def treeposition_spanning_leaves(self, start, end):
		""":returns: The tree position of the lowest descendant of this tree \
		that dominates self.leaves()[start:end].
		:raises ValueError: if end <= start """
		if end <= start:
			raise ValueError('end must be greater than start')
		# Find the tree positions of the start & end leaves,
		# and take the longest common subsequence.
		start_treepos = self.leaf_treeposition(start)
		end_treepos = self.leaf_treeposition(end - 1)
		# Find the first index where they mismatch:
		for i in range(len(start_treepos)):
			if i == len(end_treepos) or start_treepos[i] != end_treepos[i]:
				return start_treepos[:i]
		return start_treepos

	# === Convert, copy =========================================
	@classmethod
	def convert(cls, val):
		""" Convert a tree between different subtypes of Tree.

		:param cls: the class that will be used for the new tree.
		:param val: The tree that should be converted. """
		if isinstance(val, Tree):
			children = [cls.convert(child) for child in val]
			return cls(val.label, children)
		return val

	def copy(self, deep=False):
		""" Create a copy of this tree. """
		if not deep:
			return self.__class__(self.label, self)
		return self.__class__.convert(self)

	def _frozen_class(self):
		""" The frozen version of this class. """
		return ImmutableTree

	def freeze(self, leaf_freezer=None):
		""" :returns: an immutable version of this tree. """
		frozen_class = self._frozen_class()
		if leaf_freezer is None:
			newcopy = frozen_class.convert(self)
		else:
			newcopy = self.copy(deep=True)
			for pos in newcopy.treepositions('leaves'):
				newcopy[pos] = leaf_freezer(newcopy[pos])
			newcopy = frozen_class.convert(newcopy)
		hash(newcopy)  # Make sure the leaves are hashable.
		return newcopy

	# === Parsing ===============================================
	@classmethod
	def parse(cls, s, brackets='()', parse_label=None, parse_leaf=None,
			label_pattern=None, leaf_pattern=None,
			remove_empty_top_bracketing=False):
		""" Parse a bracketed tree string and return the resulting tree.
		Trees are represented as nested brackettings, such as:
		``(S (NP (NNP John)) (VP (V runs)))``

		:param s: The string to parse
		:param brackets: The two bracket characters used to mark the
			beginning and end of trees and subtrees.
		:param parse_label, parse_leaf: If specified, these functions are
			applied to the substrings of s corresponding to labels and leaves
			(respectively) to obtain the values for those labels and leaves.
			They should have the following signature: parse_label(str) -> value
		:param label_pattern, leaf_pattern: Regular expression patterns used to
			find label and leaf substrings in s. By default, both label and
			leaf patterns are defined to match any sequence of non-whitespace
			non-bracket characters.
		:param remove_empty_top_bracketing: If the resulting tree has an empty
			node label, and is length one, then return its single child
			instead. This is useful for treebank trees, which sometimes contain
			an extra level of bracketing.
		:returns: A tree corresponding to the string representation s.
			If this class method is called using a subclass of Tree, then it
			will return a tree of that type. """
		if not isinstance(brackets, basestring) or len(brackets) != 2:
			raise TypeError('brackets must be a length-2 string')
		if re.search(r'\s', brackets):
			raise TypeError('whitespace brackets not allowed')
		# Construct a regexp that will tokenize the string.
		open_b, close_b = brackets
		open_pattern, close_pattern = (re.escape(open_b), re.escape(close_b))
		if label_pattern is None:
			label_pattern = r'[^\s%s%s]+' % (open_pattern, close_pattern)
		if leaf_pattern is None:
			leaf_pattern = r'[^\s%s%s]+' % (open_pattern, close_pattern)
		token_re = re.compile(r'%s\s*(%s)?|%s|(%s)' % (
			open_pattern, label_pattern, close_pattern, leaf_pattern))
		# Walk through each token, updating a stack of trees.
		stack = [(None, [])]  # list of (label, children) tuples
		for match in token_re.finditer(s):
			token = match.group()
			if token[0] == open_b:  # Beginning of a tree/subtree
				if len(stack) == 1 and len(stack[0][1]) > 0:
					cls._parse_error(s, match, 'end-of-string')
				label = token[1:].lstrip()
				if parse_label is not None:
					label = parse_label(label)
				stack.append((label, []))
			elif token == close_b:  # End of a tree/subtree
				if len(stack) == 1:
					if len(stack[0][1]) == 0:
						cls._parse_error(s, match, open_b)
					else:
						cls._parse_error(s, match, 'end-of-string')
				label, children = stack.pop()
				stack[-1][1].append(cls(label, children))
			else:  # Leaf node
				if len(stack) == 1:
					cls._parse_error(s, match, open_b)
				if parse_leaf is not None:
					token = parse_leaf(token)
				stack[-1][1].append(token)
		# check that we got exactly one complete tree.
		if len(stack) > 1:
			cls._parse_error(s, 'end-of-string', close_b)
		elif len(stack[0][1]) == 0:
			cls._parse_error(s, 'end-of-string', open_b)
		else:
			assert stack[0][0] is None and len(stack[0][1]) == 1
		tree = stack[0][1][0]
		# If the tree has an extra level with label='', then get rid of
		# it. E.g.: "((S (NP ...) (VP ...)))"
		if remove_empty_top_bracketing and tree.label == '' and len(tree) == 1:
			tree = tree[0]
		return tree

	@classmethod
	def _parse_error(cls, s, match, expecting):
		""" Display a friendly error message when parsing a tree string fails.

		:param s: The string we're parsing.
		:param match: regexp match of the problem token.
		:param expecting: what we expected to see instead. """
		# Construct a basic error message
		if match == 'end-of-string':
			pos, token = len(s), 'end-of-string'
		else:
			pos, token = match.start(), match.group()
		msg = '%s.parse(): expected %r but got %r\n%sat index %d.' % (
			cls.__name__, expecting, token, ' ' * 12, pos)
		# Add a display showing the error token itsels:
		s = s.replace('\n', ' ').replace('\t', ' ')
		offset = pos
		if len(s) > pos + 10:
			s = s[:pos + 10] + '...'
		if pos > 10:
			s = '...' + s[pos - 10:]
			offset = 13
		msg += '\n%s"%s"\n%s^' % (' ' * 16, s, ' ' * (17 + offset))
		raise ValueError(msg)

	# === String Representation =================================
	def __repr__(self):
		childstr = ", ".join(repr(c) for c in self)
		return '%s(%r, [%s])' % (self.__class__.__name__, self.label, childstr)

	def __str__(self):
		return self._pprint_flat('', '()', False)

	def pprint(self, margin=70, indent=0, labelsep='', brackets='()',
			quotes=False):
		"""	:returns: A pretty-printed string representation of this tree.
		:param margin: The right margin at which to do line-wrapping.
		:param indent: The indentation level at which printing begins. This
			number is used to decide how far to indent subsequent lines.
		:param labelsep: A string that is used to separate the label from the
			children; e.g., the value ':' gives trees like::

				(S: (NP: I) (VP: (V: saw) (NP: it))). """
		# Try writing it on one line.
		s = self._pprint_flat(labelsep, brackets, quotes)
		if len(s) + indent < margin:
			return s
		# If it doesn't fit on one line, then write it on multi-lines.
		if isinstance(self.label, basestring):
			s = '%s%s%s' % (brackets[0], self.label, labelsep)
		else:
			s = '%s%r%s' % (brackets[0], self.label, labelsep)
		for child in self:
			if isinstance(child, Tree):
				s += '\n' + ' ' * (indent + 2) + child.pprint(margin,
						indent + 2, labelsep, brackets, quotes)
			elif isinstance(child, tuple):
				s += '\n' + ' ' * (indent + 2) + '/'.join(child)
			elif isinstance(child, basestring) and not quotes:
				s += '\n' + ' ' * (indent + 2) + '%s' % child
			else:
				s += '\n' + ' ' * (indent + 2) + '%r' % child
		return s + brackets[1]

	def _pprint_flat(self, labelsep, brackets, quotes):
		""" pretty-printing helper function. """
		childstrs = []
		for child in self:
			if isinstance(child, Tree):
				childstrs.append(child._pprint_flat(labelsep, brackets, quotes))
			elif isinstance(child, tuple):
				childstrs.append("/".join(child))
			elif isinstance(child, basestring) and not quotes:
				childstrs.append('%s' % child)
			else:
				childstrs.append('%r' % child)
		if isinstance(self.label, basestring):
			return '%s%s%s %s%s' % (brackets[0], self.label, labelsep,
									' '.join(childstrs), brackets[1])
		return '%s%r%s %s%s' % (brackets[0], self.label, labelsep,
								' '.join(childstrs), brackets[1])

	def draw(self):
		""" :returns: an ASCII art visualization of tree. """
		from discodop.treedraw import DrawTree
		return DrawTree(self, [str(a) for a in self.leaves()]).text()


class ImmutableTree(Tree):
	""" A tree which may not be modified. Has a hash() value. """
	def __init__(self, label_or_str, children=None):
		if children is None:
			return  # see note in Tree.__init__()
		super(ImmutableTree, self).__init__(label_or_str, children)
		# Precompute our hash value. This ensures that we're really
		# immutable. It also means we only have to calculate it once.
		try:
			self._hash = hash((self.label, tuple(self)))
		except (TypeError, ValueError) as err:
			raise ValueError("ImmutableTree's label and children "
					"must be immutable:\n%s %r\n%r" % (self.label, self, err))
		else:
			self._leaves = Tree.leaves(self)
			self._subtrees = tuple(Tree.subtrees(self))

	def leaves(self):
		return self._leaves

	def subtrees(self, condition=None):
		if condition is None:
			return self._subtrees
		return filter(condition, self._subtrees)

	def __setitem__(self, _index, _value):
		raise ValueError('ImmutableTrees may not be modified')

	def __setslice__(self, _start, _stop, _value):
		raise ValueError('ImmutableTrees may not be modified')

	def __delitem__(self, _index):
		raise ValueError('ImmutableTrees may not be modified')

	def __delslice__(self, _start, _stop):
		raise ValueError('ImmutableTrees may not be modified')

	def __iadd__(self):
		raise ValueError('ImmutableTrees may not be modified')

	def __imul__(self):
		raise ValueError('ImmutableTrees may not be modified')

	def append(self, _):
		raise ValueError('ImmutableTrees may not be modified')

	def extend(self, _):
		raise ValueError('ImmutableTrees may not be modified')

	def pop(self, _=None):
		raise ValueError('ImmutableTrees may not be modified')

	def remove(self, _):
		raise ValueError('ImmutableTrees may not be modified')

	def reverse(self):
		raise ValueError('ImmutableTrees may not be modified')

	def sort(self):
		raise ValueError('ImmutableTrees may not be modified')

	def __hash__(self):
		return self._hash

	def _set_label(self, label):
		""" Set self._label. This will only succeed the first time the label is
		set, which should occur in Tree.__init__(). """
		if hasattr(self, '_label'):
			raise ValueError('ImmutableTrees may not be modified')
		self._label = label  # pylint: disable=W0201

	def _get_label(self):
		""" Get node label. """
		return self._label
	label = property(_get_label, _set_label, doc=_get_label.__doc__)


class AbstractParentedTree(Tree):
	""" An abstract base class for Trees that automatically maintain pointers
	to their parents. These parent pointers are updated whenever any change is
	made to a tree's structure. Two subclasses are currently defined:

	- ParentedTree is used for tree structures where each subtree has at most
		one parent. This class should be used in cases where there is
		no "sharing" of subtrees.
	- MultiParentedTree is used for tree structures where a subtree may have
		zero or more parents. This class should be used in cases where subtrees
		may be shared.

	The AbstractParentedTree class redefines all operations that modify a
	tree's structure to call two methods, which are used by subclasses to
	update parent information:

	- ``_setparent()`` is called whenever a new child is added.
	- ``_delparent()`` is called whenever a child is removed. """
	def __init__(self, label_or_str, children=None):
		if children is None:
			return  # see note in Tree.__init__()
		super(AbstractParentedTree, self).__init__(label_or_str, children)
		# iterate over self, and *not* children, because children
		# might be an iterator.
		for i, child in enumerate(self):
			if isinstance(child, Tree):
				self._setparent(child, i, dry_run=True)
		for i, child in enumerate(self):
			if isinstance(child, Tree):
				self._setparent(child, i)

	# === Parent management =====================================
	def _setparent(self, _child, _index, dry_run=False):
		""" Update child's parent pointer to point to self. This method is only
		called if child's type is Tree; i.e., it is not called when adding a
		leaf to a tree. This method is always called before the child is
		actually added to self's child list. Typically, if child is a tree,
		then its type needs to match self's type. This prevents mixing of
		different tree types (single-, multi-, and non-parented).

		:param index: The index of child in self.
		:param dry_run: If true, the don't actually set the child's parent
			pointer; just check for any error conditions, and raise an
			exception if one is found.
		:raises TypeError: if child is a tree with an inappropriate type. """
		raise NotImplementedError('Abstract base class')

	def _delparent(self, _child, _index):
		""" Update child's parent pointer to not point to self. This method is
		only called if child's type is Tree; i.e., it is not called when
		removing a leaf from a tree. This method is always called before the
		child is actually removed from self's child list.

		:param index: The index of child in self. """
		raise NotImplementedError('Abstract base class')

	# === Methods that add/remove children ======================
	# Every method that adds or removes a child must make
	# appropriate calls to _setparent() and _delparent().

	def __delitem__(self, index):
		if isinstance(index, slice):  # del ptree[start:stop]
			start, stop = slice_bounds(self, index)
			# Clear all the children pointers.
			for i in range(start, stop):
				if isinstance(self[i], Tree):
					self._delparent(self[i], i)
			# Delete the children from our child list.
			super(AbstractParentedTree, self).__delitem__(index)
		elif isinstance(index, int):  # del ptree[i]
			if index < 0:
				index += len(self)
			if index < 0:
				raise IndexError('index out of range')
			# Clear the child's parent pointer.
			if isinstance(self[index], Tree):
				self._delparent(self[index], index)
			# Remove the child from our child list.
			super(AbstractParentedTree, self).__delitem__(index)
		elif len(index) == 0:  # del ptree[()]
			raise IndexError('The tree position () may not be deleted.')
		elif len(index) == 1:  # del ptree[(i, )]
			del self[index[0]]
		else:  # del ptree[i1, i2, i3]
			del self[index[0]][index[1:]]

	def __setitem__(self, index, value):
		if isinstance(index, slice):  # ptree[start:stop] = value
			start, stop = slice_bounds(self, index)
			# make a copy of value, in case it's an iterator
			if not isinstance(value, (list, tuple)):
				value = list(value)
			# Check for any error conditions, so we can avoid ending
			# up in an inconsistent state if an error does occur.
			for i, child in enumerate(value):
				if isinstance(child, Tree):
					self._setparent(child, start + i, dry_run=True)
			# clear the child pointers of all parents we're removing
			for i in range(start, stop):
				if isinstance(self[i], Tree):
					self._delparent(self[i], i)
			# set the child pointers of the new children. We do this
			# after clearing *all* child pointers, in case we're e.g.
			# reversing the elements in a tree.
			for i, child in enumerate(value):
				if isinstance(child, Tree):
					self._setparent(child, start + i)
			# finally, update the content of the child list itself.
			super(AbstractParentedTree, self).__setitem__(index, value)
		elif isinstance(index, int):  # ptree[i] = value
			if index < 0:
				index += len(self)
			if index < 0:
				raise IndexError('index out of range')
			# if the value is not changing, do nothing.
			if value is self[index]:
				return
			# Set the new child's parent pointer.
			if isinstance(value, Tree):
				self._setparent(value, index)
			# Remove the old child's parent pointer
			if isinstance(self[index], Tree):
				self._delparent(self[index], index)
			# Update our child list.
			super(AbstractParentedTree, self).__setitem__(index, value)
		elif len(index) == 0:  # ptree[()] = value
			raise IndexError('The tree position () may not be assigned to.')
		elif len(index) == 1:  # ptree[(i, )] = value
			self[index[0]] = value
		else:  # ptree[i1, i2, i3] = value
			self[index[0]][index[1:]] = value

	def append(self, child):
		if isinstance(child, Tree):
			self._setparent(child, len(self))
		super(AbstractParentedTree, self).append(child)

	def extend(self, children):
		for child in children:
			if isinstance(child, Tree):
				self._setparent(child, len(self))
			super(AbstractParentedTree, self).append(child)

	def insert(self, index, child):
		# Handle negative indexes. Note that if index < -len(self),
		# we do *not* raise an IndexError, unlike __getitem__. This
		# is done for consistency with list.__getitem__ and list.index.
		if index < 0:
			index += len(self)
		if index < 0:
			index = 0
		# Set the child's parent, and update our child list.
		if isinstance(child, Tree):
			self._setparent(child, index)
		super(AbstractParentedTree, self).insert(index, child)

	def pop(self, index=-1):
		if index < 0:
			index += len(self)
		if index < 0:
			raise IndexError('index out of range')
		if isinstance(self[index], Tree):
			self._delparent(self[index], index)
		return super(AbstractParentedTree, self).pop(index)

	# NB: like `list`, this is done by equality, not identity!
	# To remove a specific child, use del ptree[i].
	def remove(self, child):
		index = self.index(child)
		if isinstance(self[index], Tree):
			self._delparent(self[index], index)
		super(AbstractParentedTree, self).remove(child)

	# We need to implement __getslice__ and friends, even though
	# they're deprecated, because otherwise list.__getslice__ will get
	# called (since we're subclassing from list). Just delegate to
	# __getitem__ etc., but use max(0, start) and max(0, stop) because
	# because negative indices are already handled *before*
	# __getslice__ is called; and we don't want to double-count them.
	if hasattr(list, '__getslice__'):
		def __getslice__(self, start, stop):
			return self.__getitem__(slice(max(0, start), max(0, stop)))

		def __delslice__(self, start, stop):
			return self.__delitem__(slice(max(0, start), max(0, stop)))

		def __setslice__(self, start, stop, value):
			return self.__setitem__(slice(max(0, start), max(0, stop)), value)


class ParentedTree(AbstractParentedTree):
	""" A Tree that automatically maintains parent pointers for single-parented
	trees. The following read-only property values are automatically updated
	whenever the structure of a parented tree is modified: parent,
	parent_index, left_sibling, right_sibling, root, treeposition.
	Each ParentedTree may have at most one parent. In particular, subtrees may
	not be shared. Any attempt to reuse a single ParentedTree as a child of
	more than one parent (or as multiple children of the same parent) will
	cause a ValueError exception to be raised.
	ParentedTrees should never be used in the same tree as Trees or
	MultiParentedTrees. Mixing tree implementations may result in incorrect
	parent pointers and in TypeError exceptions. """
	def __init__(self, label_or_str, children=None):
		if children is None:
			return  # see note in Tree.__init__()
		self._parent = None
		super(ParentedTree, self).__init__(label_or_str, children)

	def _frozen_class(self):
		return ImmutableParentedTree

	# === Properties =================================================
	def _get_parent_index(self):
		""" The index of this tree in its parent;
		i.e., ptree.parent[ptree.parent_index] is ptree.
		Note that ptree.parent_index is not necessarily equal to
		ptree.parent.index(ptree), since the index() method
		returns the first child that is _equal_ to its argument. """
		if self._parent is None:
			return None
		for i, child in enumerate(self._parent):
			if child is self:
				return i
		raise ValueError('expected to find self in self._parent!')

	def _get_left_sibling(self):
		""" The left sibling of this tree, or None if it has none. """
		parent_index = self._get_parent_index()
		if self._parent and parent_index > 0:
			return self._parent[parent_index - 1]
		return None  # no left sibling

	def _get_right_sibling(self):
		""" The right sibling of this tree, or None if it has none. """
		parent_index = self._get_parent_index()
		if self._parent and parent_index < (len(self._parent) - 1):
			return self._parent[parent_index + 1]
		return None  # no right sibling

	def _get_treeposition(self):
		""" The tree position of this tree, relative to the root of the
		tree; i.e., ptree.root[ptree.treeposition] is ptree. """
		if self._parent is None:
			return ()
		return (self._parent._get_treeposition() +
				(self._get_parent_index(), ))

	def _get_root(self):
		""" The root of this tree; i.e., the unique ancestor of this tree whose
		parent is None. If ptree.parent is None, then ptree is its own root. """
		if self._parent is None:
			return self
		return self._parent._get_root()

	parent = property(lambda self: self._parent, doc="""\
		The parent of this tree, or None if it has no parent. """)
	parent_index = property(_get_parent_index, doc=_get_parent_index.__doc__)
	left_sibling = property(_get_left_sibling, doc=_get_left_sibling.__doc__)
	right_sibling = property(_get_right_sibling, doc=_get_right_sibling.__doc__)
	root = property(_get_root, doc=_get_root.__doc__)
	treeposition = property(_get_treeposition, doc=_get_treeposition.__doc__)

	# === Parent Management ==========================================
	def _delparent(self, child, index):
		assert isinstance(child, ParentedTree)
		assert self[index] is child and child._parent is self
		child._parent = None

	def _setparent(self, child, _index, dry_run=False):
		if not isinstance(child, ParentedTree):
			raise TypeError('Cannot insert a non-ParentedTree '
					'into a ParentedTree')
		if child._parent is not None:
			raise ValueError('Cannot insert a subtree that already '
					'has a parent.')
		if not dry_run:
			child._parent = self


class ImmutableParentedTree(ImmutableTree, ParentedTree):
	""" Combination of an Immutable and Parented Tree. """
	def __init__(self, label_or_str, children=None):
		if children is None:
			return  # see note in Tree.__init__()
		super(ImmutableParentedTree, self).__init__(label_or_str, children)


class MultiParentedTree(AbstractParentedTree):
	""" A Tree that automatically maintains parent pointers for multi-parented
	trees. The following read-only property values are automatically updated
	whenever the structure of a multi-parented tree is modified: parents,
	parent_indices, left_siblings, right_siblings, roots, treepositions. Each
	MultiParentedTree may have zero or more parents. In particular, subtrees
	may be shared. If a single MultiParentedTree is used as multiple children
	of the same parent, then that parent will appear multiple times in its
	parents property.
	MultiParentedTrees should never be used in the same tree as Trees or
	ParentedTrees. Mixing tree implementations may result in incorrect parent
	pointers and in TypeError exceptions. """
	def __init__(self, label_or_str, children=None):
		if children is None:
			return  # see note in Tree.__init__()
		# This list should not contain duplicates, even if a parent contains
		# this tree multiple times.
		self._parents = []
		super(MultiParentedTree, self).__init__(label_or_str, children)

	def _frozen_class(self):
		return ImmutableMultiParentedTree

	# === Properties =================================================
	def _get_parent_indices(self):
		""" Collect tuples of (parent, index) for all parents in a list. """
		return [(parent, index)
				for parent in self._parents
				for index, child in enumerate(parent)
				if child is self]

	def _get_left_siblings(self):
		""" A list of all left siblings of this tree, in any of its parent
		trees. A tree may be its own left sibling if it is used as multiple
		contiguous children of the same parent. A tree may appear multiple
		times in this list if it is the left sibling of this tree with respect
		to multiple parents. """
		return [parent[index - 1]
				for (parent, index) in self._get_parent_indices()
				if index > 0]

	def _get_right_siblings(self):
		""" A list of all right siblings of this tree, in any of its parent
		trees. A tree may be its own right sibling if it is used as multiple
		contiguous children of the same parent. A tree may appear multiple
		times in this list if it is the right sibling of this tree with respect
		to multiple parents. """
		return [parent[index + 1]
				for (parent, index) in self._get_parent_indices()
				if index < (len(parent) - 1)]

	def _get_roots(self):
		""" The set of all roots of this tree. This set is formed by tracing
		all possible parent paths until trees with no parents are found. """
		return list(self._get_roots_helper({}).values())

	def _get_roots_helper(self, result):
		""" Collect all roots for this node. """
		if self._parents:
			for parent in self._parents:
				parent._get_roots_helper(result)
		else:
			result[id(self)] = self
		return result

	parents = property(lambda self: list(self._parents), doc="""\
		The set of parents of this tree. If this tree has no parents, then
		parents is the empty set. To check if a tree is used as multiple
		children of the same parent, use the parent_indices property. """)
	left_siblings = property(_get_left_siblings, doc=_get_left_siblings.__doc__)
	right_siblings = property(_get_right_siblings,
			doc=_get_right_siblings.__doc__)
	roots = property(_get_roots, doc=_get_roots.__doc__)

	def parent_indices(self, parent):
		""" :returns: a list of the indices where this tree occurs as a child \
		of parent.

		If this child does not occur as a child of parent, then the
		empty list is returned. The following is always true::

			for parent_index in ptree.parent_indices(parent):
				parent[parent_index] is ptree """
		if parent not in self._parents:
			return []
		return [index for (index, child) in enumerate(parent)
				if child is self]

	def treepositions(self, root):
		"""	:returns: a list of all tree positions that can be used to reach \
			this multi-parented tree starting from root.

		i.e., the following holds::

			for treepos in ptree.treepositions(root):
				root[treepos] is ptree """
		if self is root:
			return [()]
		return [treepos + (index, ) for parent in self._parents
				for treepos in parent.treepositions(root)
				for (index, child) in enumerate(parent) if child is self]

	# === Parent Management ==========================================
	def _delparent(self, child, index):
		assert isinstance(child, MultiParentedTree) and self[index] is child
		assert len([p for p in child._parents if p is self]) == 1
		# If the only copy of child in self is at index, then delete
		# self from child's parent list.
		for i, c in enumerate(self):
			if c is child and i != index:
				break
		else:
			child._parents.remove(self)

	def _setparent(self, child, _index, dry_run=False):
		if not isinstance(child, MultiParentedTree):
			raise TypeError('Cannot insert a non-MultiParentedTree '
					'into a MultiParentedTree')
		# Add self as a parent pointer if it's not already listed.
		if not dry_run:
			for parent in child._parents:
				if parent is self:
					break
			else:
				child._parents.append(self)


class ImmutableMultiParentedTree(ImmutableTree, MultiParentedTree):
	""" Combination of an Immutable and Multi Parented Tree. """
	def __init__(self, label_or_str, children=None):
		if children is None:
			return  # see note in Tree.__init__()
		super(ImmutableMultiParentedTree, self).__init__(label_or_str, children)


def slice_bounds(seq, slice_obj, allow_step=False):
	""" Given a slice, return the corresponding (start, stop) bounds, taking
	into account None indices and negative indices. The following holds
	for the returned start and stop values: 0 <= start <= stop <= len(seq).

	:raises ValueError: if slice_obj.step is not None.
	:param allow_step: If true, then the slice object may have a non-None step.
		If it does, then return a tuple (start, stop, step). """
	start, stop = (slice_obj.start, slice_obj.stop)
	if allow_step:
		slice_obj.step = 1 if slice_obj.step is None else slice_obj.step
		# Use a recursive call without allow_step to find the slice
		# bounds. If step is negative, then the roles of start and
		# stop (in terms of default values, etc), are swapped.
		if slice_obj.step < 0:
			start, stop = slice_bounds(seq, slice(stop, start))
		else:
			start, stop = slice_bounds(seq, slice(start, stop))
		return start, stop, slice_obj.step
	elif slice_obj.step not in (None, 1):
		raise ValueError('slices with steps are not supported by %s' %
				seq.__class__.__name__)
	start = 0 if start is None else start
	stop = len(seq) if stop is None else stop
	start = max(0, len(seq) + start) if start < 0 else start
	stop = max(0, len(seq) + stop) if stop < 0 else stop
	if stop > 0:  # Make sure stop doesn't go past the end of the list.
		# Try to avoid calculating len(seq), since that can be expensive for
		# lazy sequences.
		try:
			seq[stop - 1]
		except IndexError:
			stop = len(seq)
	start = min(start, stop)
	return start, stop



""" Simple command line interface to parse with grammar(s) in text format.  """
from __future__ import print_function
import io
import os
import re
import sys
import time
import gzip
import codecs
import logging
import tempfile
import string  # pylint: disable=W0402
from math import log
from getopt import gnu_getopt, GetoptError
from operator import itemgetter
from itertools import count
from subprocess import Popen, PIPE
import numpy as np
from discodop import plcfrs, pcfg
from discodop.grammar import FORMAT, defaultparse
from discodop._grammar import Grammar
from discodop.containers import CFGChartItem
from discodop.coarsetofine import prunechart, whitelistfromposteriors
from discodop.disambiguation import marginalize, doprerank
from discodop.tree import Tree
from discodop.lexicon import replaceraretestwords, getunknownwordfun, UNK
from discodop.treebank import saveheads, TERMINALSRE
from discodop.treebanktransforms import reversetransform, rrbacktransform
from discodop.treetransforms import mergediscnodes, unbinarize, \
		removefanoutmarkers

USAGE = """
usage: %s [options] <rules> <lexicon> [input [output]]
or:    %s [options] --ctf k <coarserules> <coarselex>
          <finerules> <finelex> [input [output]]

Grammars need to be binarized, and are in bitpar or PLCFRS format.
When no file is given, output is written to standard output;
when additionally no input is given, it is read from standard input.
Files must be encoded in UTF-8.
Input should contain one token per line, with sentences delimited by two
newlines. Output consists of bracketed trees, with discontinuities indicated
through indices pointing to words in the original sentence.

Options:
  -b k          Return the k-best parses instead of just 1.
  -s x          Use "x" as start symbol instead of default "TOP".
  -z            Input is one sentence per line, space-separated tokens.
  --ctf=k       Use k-best coarse-to-fine; prune items not in top k derivations
  --prob        Print probabilities as well as parse trees.
  --mpd         In coarse-to-fine mode, produce the most probable
                derivation (MPD) instead of the most probable parse (MPP).

%s
""" % (sys.argv[0], sys.argv[0], FORMAT)

DEFAULTSTAGE = dict(
		name='stage1',  # identifier, used for filenames
		mode='plcfrs',  # use the agenda-based PLCFRS parser
		prune=False,  # whether to use previous chart to prune this stage
		split=False,  # split disc. nodes VP_2[101] as { VP*[100], VP*[001] }
		splitprune=False,  # treat VP_2[101] as {VP*[100], VP*[001]} for pruning
		markorigin=False,  # mark origin of split nodes: VP_2 => {VP*1, VP*2}
		k=50,  # no. of coarse pcfg derivations to prune with; k=0: filter only
		neverblockre=None,  # do not prune nodes with label that match regex
		getestimates=None,  # compute & store estimates
		useestimates=None,  # load & use estimates
		dop=False,  # enable DOP mode (DOP reduction / double DOP)
		packedgraph=False,  # use packed graph encoding for DOP reduction
		usedoubledop=False,  # when False, use DOP reduction instead
		iterate=False,  # for double dop, whether to add fragments of fragments
		complement=False,  # for double dop, whether to include fragments which
			# form the complement of the maximal recurring fragments extracted
		sample=False, kbest=True,
		m=10,  # number of derivations to sample/enumerate
		estimator='ewe',  # choices: dop1, ewe
		objective='mpp',  # choices: mpp, mpd, shortest, sl-dop[-simple]
			# NB: w/shortest derivation, estimator only affects tie breaking.
		sldop_n=7)


def main():
	""" Handle command line arguments. """
	print('PLCFRS parser - Andreas van Cranenburgh', file=sys.stderr)
	options = 'ctf= prob mpd'.split()
	try:
		opts, args = gnu_getopt(sys.argv[1:], 'u:b:s:z', options)
		assert 2 <= len(args) <= 6, 'incorrect number of arguments'
	except (GetoptError, AssertionError) as err:
		print(err, USAGE)
		return
	for n, filename in enumerate(args):
		assert os.path.exists(filename), (
				'file %d not found: %r' % (n + 1, filename))
	opts = dict(opts)
	k = int(opts.get('-b', 1))
	top = opts.get('-s', 'TOP')
	threshold = int(opts.get('--ctf', 0))
	prob = '--prob' in opts
	oneline = '-z' in opts
	rules = (gzip.open if args[0].endswith('.gz') else open)(args[0]).read()
	lexicon = codecs.getreader('utf-8')((gzip.open if args[1].endswith('.gz')
			else open)(args[1])).read()
	bitpar = rules[0] in string.digits
	coarse = Grammar(rules, lexicon, start=top, bitpar=bitpar)
	stages = []
	stage = DEFAULTSTAGE.copy()
	stage.update(
			name='coarse',
			mode='pcfg' if bitpar else 'plcfrs',
			grammar=coarse,
			backtransform=None,
			m=k)
	stages.append(DictObj(stage))
	if 4 <= len(args) <= 6 and threshold:
		rules = (gzip.open if args[2].endswith('.gz') else open)(args[2]).read()
		lexicon = codecs.getreader('utf-8')((gzip.open
				if args[3].endswith('.gz') else open)(args[3])).read()
		# detect bitpar format
		bitpar = rules[0] in string.digits
		fine = Grammar(rules, lexicon, start=top, bitpar=bitpar)
		fine.getmapping(coarse, striplabelre=re.compile(b'@.+$'))
		stage = DEFAULTSTAGE.copy()
		stage.update(
				name='fine',
				mode='pcfg' if bitpar else 'plcfrs',
				grammar=fine,
				backtransform=None,
				m=k,
				prune=True,
				k=threshold,
				objective='mpd' if '--mpd' in opts else 'mpp')
		stages.append(DictObj(stage))
		infile = (io.open(args[4], encoding='utf-8')
				if len(args) >= 5 else sys.stdin)
		out = (io.open(args[5], 'w', encoding='utf-8')
				if len(args) == 6 else sys.stdout)
	else:
		infile = (io.open(args[2], encoding='utf-8')
				if len(args) >= 3 else sys.stdin)
		out = (io.open(args[3], 'w', encoding='utf-8')
				if len(args) == 4 else sys.stdout)
	doparsing(Parser(stages), infile, out, prob, oneline)


def doparsing(parser, infile, out, printprob, oneline):
	""" Parse sentences from file and write results to file, log to stdout. """
	times = [time.clock()]
	unparsed = 0
	if not oneline:
		infile = infile.read().split('\n\n')
	for n, line in enumerate(infile):
		if not line.strip():
			continue
		sent = line.split() if oneline else line.splitlines()
		lexicon = parser.stages[0].grammar.lexicalbyword
		assert not set(sent) - set(lexicon), (
			'unknown words and no open class tags supplied: %r' % (
			list(set(sent) - set(lexicon))))
		print('parsing %d: %s' % (n, ' '.join(sent)), file=sys.stderr)
		sys.stdout.flush()
		result = list(parser.parse(sent))[-1]
		if result.noparse:
			unparsed += 1
			out.writelines('No parse for "%s"\n' % ' '.join(sent))
		elif printprob:
			out.writelines('prob=%.16g\n%s\n' % (prob, tree)
					for tree, prob in sorted(result.parsetrees.items(),
						key=itemgetter(1), reverse=True))
		else:
			out.writelines('%s\n' % tree
					for tree in sorted(result.parsetrees,
						key=result.parsetrees.get, reverse=True))
		out.flush()
		times.append(time.clock())
		print(times[-1] - times[-2], 's', file=sys.stderr)
	times = [a - b for a, b in zip(times[1::2], times[::2])]
	print('raw cpu time', time.clock() - times[0],
			'\naverage time per sentence', sum(times) / len(times),
			'\nunparsed sentences:', unparsed,
			'\nfinished',
			file=sys.stderr)
	out.close()


class Parser(object):
	""" An object to parse sentences following parameters given as a sequence
	of coarse-to-fine stages.

	:param stages: a list of coarse-to-fine stages containing grammars and
		parameters.
	:param transformations: treebank transformations to reverse on parses.
	:param tailmarker: if heads have been marked with a symbol, use this to
		mark heads in the output.
	:param postagging: if given, an unknown word model is used, consisting of a
		dictionary with three items:
		- unknownwordfun: function to produces signatures for unknown words.
		- lexicon: the set of known words in the grammar.
		- sigs: the set of word signatures occurring in the grammar.
	:param relationalrealizational: whether to reverse the RR-transform. """
	def __init__(self, stages, transformations=None, tailmarker=None,
			relationalrealizational=None, postagging=None):
		self.stages = stages
		self.transformations = transformations
		self.tailmarker = tailmarker
		self.postagging = postagging
		self.relationalrealizational = relationalrealizational

	def parse(self, sent, tags=None):
		""" Parse a sentence and yield a dictionary from parse trees to
		probabilities for each stage.

		:param tags: if given, will be given to the parser instead of trying
			all possible tags. """
		if self.postagging:
			sent = replaceraretestwords(sent,
					self.postagging['unknownwordfun'],
					self.postagging['lexicon'], self.postagging['sigs'])
		sent = list(sent)
		if tags is not None:
			tags = list(tags)
		chart = {}
		start = inside = outside = None
		for n, stage in enumerate(self.stages):
			begin = time.clock()
			noparse = False
			parsetrees = fragments = None
			msg = '%s:\t' % stage.name.upper()
			model = u'default'
			if stage.dop:
				if (stage.estimator == 'ewe'
						or stage.objective.startswith('sl-dop')):
					model = u'ewe'
				elif stage.estimator == 'bon':
					model = u'bon'
				if stage.objective == 'shortest':
					model = u'shortest'
			x = stage.grammar.currentmodel
			stage.grammar.switch(model, logprob=stage.mode != 'pcfg-posterior')
			if stage.mode == 'pcfg-bitpar' and (
					not hasattr(stage, 'rulesfile')
					or x != stage.grammar.currentmodel):
				exportbitpargrammar(stage)
			if not stage.prune or start:
				if n != 0 and stage.prune and stage.mode != 'dop-rerank':
					if self.stages[n - 1].mode == 'pcfg-posterior':
						(whitelist, sentprob, unfiltered,
							numitems, numremain) = whitelistfromposteriors(
								inside, outside, start,
								self.stages[n - 1].grammar, stage.grammar,
								stage.k, stage.splitprune,
								self.stages[n - 1].markorigin,
								stage.mode.startswith('pcfg'))
						msg += ('coarse items before pruning=%d; filtered: %d;'
								' pruned: %d; sentprob=%g\n\t' % (
								unfiltered, numitems, numremain, sentprob))
					else:
						whitelist, items = prunechart(
								chart, start, self.stages[n - 1].grammar,
								stage.grammar, stage.k, stage.splitprune,
								self.stages[n - 1].markorigin,
								stage.mode.startswith('pcfg'),
								self.stages[n - 1].mode == 'pcfg-bitpar')
						msg += 'coarse items before pruning: %d; ' % (
								sum(len(a) for x in chart for a in x if a)
								if self.stages[n - 1].mode == 'pcfg'
								else len(chart))
						msg += 'after: %d\n\t' % (items)
				else:
					whitelist = None
				if stage.mode == 'pcfg':
					chart, start, msg1 = pcfg.parse(
							sent, stage.grammar, tags=tags,
							chart=whitelist if stage.prune else None)
				elif stage.mode == 'pcfg-symbolic':
					chart, start, msg1 = pcfg.parse_symbolic(
							sent, stage.grammar, tags=tags)
				elif stage.mode == 'pcfg-posterior':
					inside, outside, start, msg1 = pcfg.doinsideoutside(
							sent, stage.grammar, tags=tags)
				elif stage.mode == 'pcfg-bitpar':
					chart, start, msg1 = parse_bitpar(
							stage.rulesfile.name, stage.lexiconfile.name,
							sent, stage.m, stage.grammar.start,
							stage.grammar.toid[stage.grammar.start], tags=tags)
					msg1 += '%d derivations' % (len(chart) if start else 0)
				elif stage.mode == 'plcfrs':
					chart, start, msg1 = plcfrs.parse(
							sent, stage.grammar, tags=tags,
							exhaustive=stage.dop or (n + 1 != len(self.stages)
								and self.stages[n + 1].prune),
							whitelist=whitelist,
							splitprune=stage.splitprune
								and self.stages[n - 1].split,
							markorigin=self.stages[n - 1].markorigin,
							estimates=(stage.useestimates, stage.outside)
								if stage.useestimates in ('SX', 'SXlrgaps')
								else None)
				elif stage.mode == 'dop-rerank':
					if start:
						parsetrees = doprerank(chart, start, sent, stage.k,
								self.stages[n - 1].grammar, stage.grammar)
						msg1 = 're-ranked %d parse trees. ' % len(parsetrees)
				else:
					raise ValueError('unknown mode specified.')
				msg += '%s\n\t' % msg1
				if (n != 0 and not start and not noparse
						and stage.split == self.stages[n - 1].split):
					logging.error('ERROR: expected successful parse. '
							'sent: %s\nstage: %s.', ' '.join(sent), stage.name)
					#raise ValueError('ERROR: expected successful parse. '
					#		'sent %s, %s.' % (nsent, stage.name))
			if start and stage.mode not in ('pcfg-posterior', 'dop-rerank'
					) and not (self.relationalrealizational and stage.split):
				begindisamb = time.clock()
				if stage.objective == 'shortest':
					stage.grammar.switch('ewe' if stage.estimator == 'ewe'
							else 'default', True)
				parsetrees, derivs, msg1 = marginalize(stage.objective
						if stage.dop else 'mpd',
						chart, start, stage.grammar, stage.m,
						sample=stage.sample, kbest=stage.kbest,
						sent=sent, tags=tags,
						sldop_n=stage.sldop_n,
						backtransform=stage.backtransform,
						bitpar=stage.mode == 'pcfg-bitpar')
				msg += 'disambiguation: %s, %gs\n\t' % (
						msg1, time.clock() - begindisamb)
			if parsetrees:
				resultstr, prob = max(parsetrees.items(), key=itemgetter(1))
				try:
					parsetree, fragments, noparse = self.postprocess(
							resultstr, n, derivs)
				except ValueError as err:
					logging.error("something's amiss: %r", err)
					parsetree, prob, fragments, noparse = self.noparse(
							stage, sent, tags)
				msg += probstr(prob) + ' '
			else:
				parsetree, prob, fragments, noparse = self.noparse(
						stage, sent, tags)
			elapsedtime = time.clock() - begin
			msg += '%.2fs cpu time elapsed\n' % (elapsedtime)
			yield DictObj(name=stage.name, parsetree=parsetree, prob=prob,
					parsetrees=parsetrees, fragments=fragments,
					noparse=noparse, elapsedtime=elapsedtime, msg=msg)

	def postprocess(self, treestr, stage=-1, derivs=None):
		""" Take parse tree and apply postprocessing. """
		parsetree = Tree.parse(treestr, parse_leaf=int)
		if self.stages[stage].split:
			mergediscnodes(unbinarize(parsetree, childchar=':'))
		saveheads(parsetree, self.tailmarker)
		unbinarize(parsetree)
		removefanoutmarkers(parsetree)
		if self.relationalrealizational:
			parsetree = rrbacktransform(parsetree,
					self.relationalrealizational['adjunctionlabel'])
		if self.transformations:
			reversetransform(parsetree, self.transformations)
		fragments = derivs.get(treestr) if derivs else None
		return parsetree, fragments, False

	def noparse(self, stage, sent, tags):
		""" Produce a dummy parse for evaluation purposes. """
		parsetree = defaultparse([(n, t)
				for n, t in enumerate(tags or (len(sent) * ['NONE']))])
		parsetree = Tree.parse('(%s %s)' % (stage.grammar.start,
				parsetree), parse_leaf=int)
		return parsetree, 0.0, None, True


def readgrammars(resultdir, stages, postagging=None, top='ROOT'):
	""" Read the grammars from a previous experiment.
	Expects a directory 'resultdir' which contains the relevant grammars and
	the parameter file 'params.prm', as produced by runexp. """
	for n, stage in enumerate(stages):
		logging.info('reading: %s', stage.name)
		rules = gzip.open('%s/%s.rules.gz' % (resultdir, stage.name))
		lexicon = codecs.getreader('utf-8')(gzip.open('%s/%s.lex.gz' % (
				resultdir, stage.name)))
		grammar = Grammar(rules.read(), lexicon.read(),
				start=top, bitpar=stage.mode.startswith('pcfg'))
		backtransform = None
		if stage.dop:
			assert stage.useestimates is None, 'not supported'
			if stage.usedoubledop:
				backtransform = gzip.open('%s/%s.backtransform.gz' % (
						resultdir, stage.name)).read().splitlines()
				if n and stage.prune:
					_ = grammar.getmapping(stages[n - 1].grammar,
						striplabelre=re.compile(b'@.+$'),
						neverblockre=re.compile(b'^#[0-9]+|.+}<'),
						splitprune=stage.splitprune and stages[n - 1].split,
						markorigin=stages[n - 1].markorigin)
				else:
					# recoverfragments() relies on this mapping to identify
					# binarization nodes
					_ = grammar.getmapping(None,
						neverblockre=re.compile(b'.+}<'))
			elif n and stage.prune:  # dop reduction
				_ = grammar.getmapping(stages[n - 1].grammar,
					striplabelre=re.compile(b'@[-0-9]+$'),
					neverblockre=re.compile(stage.neverblockre)
						if stage.neverblockre else None,
					splitprune=stage.splitprune and stages[n - 1].split,
					markorigin=stages[n - 1].markorigin)
				if stage.mode == 'dop-rerank':
					grammar.getrulemapping(stages[n - 1].grammar)
			probmodels = np.load('%s/%s.probs.npz' % (resultdir, stage.name))
			for name in probmodels.files:
				if name != 'default':
					grammar.register(unicode(name), probmodels[name])
		else:  # not stage.dop
			if n and stage.prune:
				_ = grammar.getmapping(stages[n - 1].grammar,
					neverblockre=re.compile(stage.neverblockre)
						if stage.neverblockre else None,
					splitprune=stage.splitprune and stages[n - 1].split,
					markorigin=stages[n - 1].markorigin)
		if stage.mode == 'pcfg-bitpar':
			assert grammar.maxfanout == 1
		grammar.testgrammar()
		stage.update(grammar=grammar, backtransform=backtransform, outside=None)
	if postagging and postagging['method'] == 'unknownword':
		postagging['unknownwordfun'] = getunknownwordfun(postagging['model'])
		postagging['lexicon'] = {w for w in stages[0].grammar.lexicalbyword
				if not w.startswith(UNK)}
		postagging['sigs'] = {w for w in stages[0].grammar.lexicalbyword
				if w.startswith(UNK)}


def exportbitpargrammar(stage):
	""" (re-)export bitpar grammar with current weights. """
	if not hasattr(stage, 'rulesfile'):
		stage.rulesfile = tempfile.NamedTemporaryFile()
		stage.lexiconfile = tempfile.NamedTemporaryFile()
	stage.rulesfile.seek(0)
	stage.rulesfile.truncate()
	if stage.grammar.currentmodel == 0:
		stage.rulesfile.write(stage.grammar.origrules)
	else:
		stage.rulesfile.writelines(
				'%g\t%s\n' % (weight, line.split(None, 1)[1])
				for weight, line in
				zip(stage.grammar.models[stage.grammar.currentmodel],
					stage.grammar.origrules.splitlines()))
	stage.rulesfile.flush()

	stage.lexiconfile.seek(0)
	stage.lexiconfile.truncate()
	lexicon = stage.grammar.origlexicon.replace(
			'(', '-LRB-').replace(')', '-RRB-')
	lexiconfile = codecs.getwriter('utf-8')(stage.lexiconfile)
	if stage.grammar.currentmodel == 0:
		lexiconfile.writelines(lexicon)
	else:
		weights = iter(stage.grammar.models[stage.grammar.currentmodel,
				stage.grammar.numrules:])
		lexiconfile.writelines('%s\t%s\n' % (line.split(None, 1)[0],
				'\t'.join('%s %g' % (tag, next(weights))
					for tag in line.split()[1::2]))
				for line in lexicon.splitlines())
	stage.lexiconfile.flush()


class DictObj(object):
	""" Trivial class to wrap a dictionary for reasons of syntactic sugar. """

	def __init__(self, *a, **kw):
		self.__dict__.update(*a, **kw)

	def update(self, *a, **kw):
		""" Update/add more attributes. """
		self.__dict__.update(*a, **kw)

	def __getattr__(self, name):
		""" This is only called when the normal mechanism fails, so in practice
		should never be called. It is only provided to satisfy pylint that it
		is okay not to raise E1101 errors in the client code. """
		raise AttributeError('%r instance has no attribute %r' % (
				self.__class__.__name__, name))

	def __repr__(self):
		return '%s(%s)' % (self.__class__.__name__,
			',\n'.join('%s=%r' % a for a in self.__dict__.items()))


def probstr(prob):
	""" Render probability / number of subtrees as string. """
	if isinstance(prob, tuple):
		return 'subtrees=%d, p=%.4g ' % (abs(prob[0]), prob[1])
	return 'p=%.4g' % prob


BITPARUNESCAPE = re.compile(r"\\([#{}\[\]<>\^$'])")
BITPARPARSES = re.compile(r'(?:^|\n)vitprob=(.*)\n(\(.*\))\n')
BITPARPARSESLOG = re.compile(r'(?:^|\n)logvitprob=(.*)\n(\(.*\))\n')


def parse_bitpar(rulesfile, lexiconfile, sent, n,
		startlabel, startid, tags=None):
	""" Parse a single sentence with bitpar, given filenames of rules and
	lexicon. n is the number of derivations to ask for (max 1000).
	Result is a dictionary of derivations with their probabilities. """
	# TODO: get full viterbi parse forest, turn into chart w/ChartItems
	assert 1 <= n <= 1000
	if tags:
		tmp = tempfile.NamedTemporaryFile()
		tmp.writelines('%s\t%s 1\n' % (t, t) for t in set(tags))
		lexiconfile = tmp.name
	tokens = [word.replace('(', '-LRB-').replace(')', '-RRB-').encode('utf8')
			for word in (tags or sent)]
	proc = Popen(['bitpar', '-q', '-vp', '-b', str(n), '-s', startlabel,
			rulesfile, lexiconfile],
			shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE)
	results, msg = proc.communicate('\n'.join(tokens) + '\n')
	# decode results or not?
	if not results or results.startswith('No parse'):
		return {}, None, '%s\n%s' % (results, msg)
	start = CFGChartItem(startid, 0, len(sent))
	lines = BITPARUNESCAPE.sub(r'\1', results).replace(')(', ') (')
	derivs = {renumber(deriv): -float(prob)
			for prob, deriv in BITPARPARSESLOG.findall(lines)}
	if not derivs:
		derivs = {renumber(deriv): -log(float(prob))
				for prob, deriv in BITPARPARSES.findall(lines)}
	return derivs, start, msg


def renumber(deriv):
	""" Replace terminals of CF-derivation (string) with indices. """
	it = count()
	return TERMINALSRE.sub(lambda _: ' %s)' % next(it), deriv)


if __name__ == '__main__':
	main()



""" Generate random sentences with an LCFRS. Reads grammar from a
text file. """
from __future__ import print_function
import sys
import gzip
import codecs
from collections import namedtuple, defaultdict
from array import array
from random import random

USAGE = """
Generate random sentences with a PLCFRS or PCFG.
Reads grammar from a text file in PLCFRS or bitpar format.
Usage: %s [--verbose] <rules> <lexicon>
or: %s --test

Grammar is assumed to be in UTF-8; may be gzip'ed (.gz extension).
""" % (sys.argv[0], sys.argv[0])

Grammar = namedtuple("Grammar", ('numrules', 'unary', 'lbinary', 'rbinary',
		'bylhs', 'lexicalbyword', 'lexicalbylhs', 'toid', 'tolabel', 'fanout'))
Rule = namedtuple("Rule",
		('lhs', 'rhs1', 'rhs2', 'args', 'lengths', 'prob', 'no'))
LexicalRule = namedtuple("LexicalRule",
		('lhs', 'rhs1', 'rhs2', 'word', 'prob', 'no'))


def gen(grammar, start=1, discount=0.75, prodcounts=None, verbose=False):
	""" generate a random sentence in top-down fashion.
	discount is a factor between 0 and 1.0; 1.0 means no discount,
	lower values introduce increasingly larger discount for repeated rules.
	Cf. http://eli.thegreenplace.net/2010/01/28/generating-random-sentences\
	-from-a-context-free-grammar/
	"""
	if prodcounts is None:
		prodcounts = [1] * grammar.numrules
	if not grammar.bylhs[start]:
		terminal = chooserule(grammar.lexicalbylhs[start], discount, prodcounts)
		return (terminal.prob, [[terminal.word]])
	rule = chooserule(grammar.bylhs[start], discount, prodcounts)
	prodcounts[rule.no] += 1
	p1, l1 = gen(grammar, rule.rhs1, discount, prodcounts, verbose)
	assert l1
	if rule.rhs2:
		p2, l2 = gen(grammar, rule.rhs2, discount, prodcounts, verbose)
		assert l2
	prodcounts[rule.no] -= 1
	if verbose:
		print('%s => %s %r [p=%s]' % (
				grammar.tolabel[rule.lhs],
				' '.join((grammar.tolabel[rule.rhs1],
				grammar.tolabel[rule.rhs2] if rule.rhs2 else '')),
				arraytoyf(rule.args, rule.lengths), rule.prob))
	if rule.rhs2:
		return compose(rule, (p1, l1), (p2, l2), verbose)
	return (p1 * rule.prob, l1)


def chooserule(rules, discount, prodcounts):
	""" given a list of objects with probabilities,
	choose one according to that distribution. """
	weights = [rule.prob * discount ** prodcounts[rule.no] for rule in rules]
	position = random() * sum(weights)
	for r, w in zip(rules, weights):
		position -= w
		if position < 0:
			return r
	raise ValueError


def compose(rule, left, right, verbose):
	""" Combine the results of two generated non-terminals into a single
	non-terminal, as specified by the given rule. """
	(p1, l1), (p2, l2) = left, right
	result = []
	if verbose:
		print("[%s] %s + %s = " % (rule.prob * p1 * p2, l1, l2), end='')
	for n, a in enumerate(rule.lengths):
		arg = []
		for b in range(a):
			if (rule.args[n] >> b) & 1:
				arg += l2.pop(0)
			else:
				arg += l1.pop(0)
		result.append(arg)
	if verbose:
		print(result)
	return (rule.prob * p1 * p2, result)


def parsefrac(a):
	""" Parse a string of a fraction into a float ('1/2' => 0.5).
	Substitute for creating Fraction objects (which is slow). """
	n = a.find('/')
	if n == -1:
		return float(a)
	return float(a[:n]) / float(a[n + 1:])


def read_lcfrs_grammar(rules, lexicon):
	""" Reads a grammar as produced by write_lcfrs_grammar from two file
	objects. """
	rules = (a.strip().split('\t') for a in rules)
	grammar = [((tuple(a[:-2]), tuple(tuple(map(int, b))
			for b in a[-2].split(","))), parsefrac(a[-1])) for a in rules]
	# one word per line, word (tag weight)+
	grammar += [(((t, 'Epsilon'), (lexentry[0],)), parsefrac(p))
			for lexentry in (a.strip().split() for a in lexicon)
			for t, p in zip(lexentry[1::2], lexentry[2::2])]
	return grammar


def read_bitpar_grammar(rules, lexicon):
	""" Read a bitpar grammar given two file objects. Must be a binarized
	grammar. Integer frequencies will be converted to exact relative
	frequencies; otherwise weights are kept as-is. """
	grammar = []
	integralweights = True
	ntfd = defaultdict(int)
	for a in rules:
		a = a.split()
		p, rule = float(a[0]), a[1:]
		if integralweights:
			ip = int(p)
			if p == ip:
				p = ip
			else:
				integralweights = False
		ntfd[rule[0]] += p
		if len(rule) == 2:
			grammar.append(((tuple(rule), ((0,),)), p))
		elif len(rule) == 3:
			grammar.append(((tuple(rule), ((0, 1),)), p))
		else:
			raise ValueError("grammar is not binarized")
	for a in lexicon:
		a = a.split()
		word = a[0]
		tags, weights = a[1::2], a[2::2]
		weights = map(float, weights)
		if integralweights:
			if all(int(w) == w for w in weights):
				weights = map(int, weights)
			else:
				integralweights = False
		tags = zip(tags, weights)
		for t, p in tags:
			ntfd[t] += p
		grammar.extend((((t, 'Epsilon'), (word,)), p) for t, p in tags)
	if integralweights:
		return [(rule, p / ntfd[rule[0][0]]) for rule, p in grammar]
	return grammar


def splitgrammar(rules):
	""" split the grammar into various lookup tables, mapping nonterminal
	labels to numeric identifiers. Also turns probabilities into negative
	log-probabilities. Can only represent binary, monotone LCFRS rules. """
	# get a list of all nonterminals; make sure Epsilon and ROOT are first,
	# and assign them unique IDs
	nonterminals = list(enumerate(["Epsilon", "ROOT"]
		+ sorted(set(str(nt) for (rule, yf), weight in rules for nt in rule)
			- {"Epsilon", "ROOT"})))
	grammar = Grammar(
			toid=dict((lhs, n) for n, lhs in nonterminals),
			tolabel=dict((n, lhs) for n, lhs in nonterminals),
			bylhs=[[] for _ in nonterminals],
			unary=[[] for _ in nonterminals],
			lbinary=[[] for _ in nonterminals],
			rbinary=[[] for _ in nonterminals],
			fanout=array('B', [0] * len(nonterminals)),
			lexicalbyword={},
			lexicalbylhs={},
			numrules=len(rules))
	for n, ((rule, yf), w) in enumerate(rules):
		if rule[1] == 'Epsilon':
			word = yf[0]
			t = LexicalRule(grammar.toid[rule[0]], 0, 0, word, w, n)
			assert grammar.fanout[t.lhs] in (0, 1)
			grammar.fanout[t.lhs] = 1
			grammar.lexicalbyword.setdefault(word, []).append(t)
			grammar.lexicalbylhs.setdefault(t.lhs, []).append(t)
			continue
		else:
			args, lengths = yfarray(yf)
			assert yf == arraytoyf(args, lengths), "rule not binarized?"
			if len(rule) == 2 and w == 1:
				w -= 0.00000001
			r = Rule(grammar.toid[rule[0]], grammar.toid[rule[1]],
					grammar.toid[rule[2]] if len(rule) == 3 else 0, args,
					lengths, w, n)
			if grammar.fanout[r.lhs] == 0:
				grammar.fanout[r.lhs] = len(args)
			assert grammar.fanout[r.lhs] == len(args)
		if len(rule) == 2:
			grammar.unary[r.rhs1].append(r)
			grammar.bylhs[r.lhs].append(r)
		elif len(rule) == 3:
			grammar.lbinary[r.rhs1].append(r)
			grammar.rbinary[r.rhs2].append(r)
			grammar.bylhs[r.lhs].append(r)
		else:
			raise ValueError("grammar not binarized: %s" % repr(r))
	#assert 0 not in grammar.fanout[1:]
	return grammar


def yfarray(yf):
	""" convert a yield function represented as a 2D sequence to an array
	object. """
	# I for 32 bits (int), H for 16 bits (short), B for 8 bits (char)
	# obviously, all related static declarations should match these types
	lentype = 'H'
	vectype = 'I'
	lensize = 8 * array(lentype).itemsize
	vecsize = 8 * array(vectype).itemsize
	assert len(yf) <= lensize
	assert all(len(a) <= vecsize for a in yf)
	initializer = [sum(2 ** n * b for n, b in enumerate(a)) for a in yf]
	return array(vectype, initializer), array(lentype, map(len, yf))


def arraytoyf(args, lengths):
	""" Inverse of yfarray(). """
	return tuple(tuple(1 if a & (1 << m) else 0 for m in range(n))
							for n, a in zip(lengths, args))


def test():
	""" Demonstration on an example grammar. """
	rules = [
		((('S', 'VP2', 'VMFIN'), ((0, 1, 0), )), 1),
		((('VP2', 'VP2', 'VAINF'), ((0, ), (0, 1))), 1. / 2),
		((('VP2', 'PROAV', 'VVPP'), ((0, ), (1, ))), 1. / 2),
		((('VP2', 'VP2'), ((0, ), (0, ))), 1. / 10),
		((('PROAV', 'Epsilon'), ('Darueber', )), 1),
		((('VAINF', 'Epsilon'), ('werden', )), 1),
		((('VMFIN', 'Epsilon'), ('muss', )), 1),
		((('VVPP', 'Epsilon'), ('nachgedacht', )), 1)]
	grammar = splitgrammar(rules)
	_, sent = gen(grammar, start=grammar.toid['S'], verbose=True)
	print(' '.join(sent.pop()))


def main():
	""" Load a grammar from a text file and generate 20 sentences. """
	if "--test" in sys.argv:
		test()
		return
	start = "ROOT"
	if "-s" in sys.argv:
		i = sys.argv.index("-s")
		start = sys.argv.pop(i + 1)
		sys.argv[i:i + 2] = []
	verbose = "--verbose" in sys.argv
	if verbose:
		sys.argv.remove('--verbose')
	if len(sys.argv) != 3:
		print(USAGE)
		return
	rules = (gzip.open if sys.argv[1].endswith(".gz") else open)(sys.argv[1])
	lexicon = codecs.getreader('utf-8')((gzip.open
			if sys.argv[2].endswith(".gz") else open)(sys.argv[2]))
	try:
		xgrammar = read_lcfrs_grammar(rules, lexicon)
	except ValueError:
		xgrammar = read_bitpar_grammar(rules, lexicon)
	grammar = splitgrammar(xgrammar)
	for _ in range(20):
		p, sent = gen(grammar, start=grammar.toid[start], verbose=verbose)
		print("[%g] %s" % (p, ' '.join(sent.pop())))

if __name__ == '__main__':
	main()



""" Web interface to draw trees. Requires Flask.
Optional: pdflatex, tikz, imagemagick. """
# stdlib
import os
import re
from subprocess import Popen, PIPE
# Flask & co
from flask import Flask, Response
from flask import request, render_template, redirect, url_for
from flask import send_from_directory
# disco-dop
from discodop.treebank import incrementaltreereader
from discodop.treedraw import DrawTree


LIMIT = 1024 * 10  # ~10KB
MORPH_TAGS = re.compile(
		r'\(([_*A-Z0-9]+)(?:\[[^ ]*\][0-9]?)?((?:-[_A-Z0-9]+)?(?:\*[0-9]+)? )')
FUNC_TAGS = re.compile(r'-[_A-Z0-9]+')
GETLEAVES = re.compile(r' ([^ ()]+)(?=[ )])')
PREAMBLE = r"""\documentclass{article}
\usepackage[landscape]{geometry}
\usepackage[utf8]{inputenc}
%s
%% NB: preview is optional, to make a cropped pdf
\usepackage[active, tightpage]{preview} \setlength{\PreviewBorder}{0.2cm}
\begin{document}
\pagestyle{empty}
\fontfamily{phv}\selectfont
\begin{preview}
"""
POSTAMBLE = r"""
\end{preview}
\end{document}
"""
APP = Flask(__name__)


@APP.route('/')
def main():
	""" Redirect to avoid trailing slash hassles. """
	return redirect(url_for('index'))


@APP.route('/index')
def index():
	""" Form for trees & parameters. """
	return render_template('draw.html')


@APP.route('/favicon.ico')
def favicon():
	""" Serve the favicon. """
	return send_from_directory(os.path.join(APP.root_path, 'static'),
			'treesearch.ico', mimetype='image/vnd.microsoft.icon')


@APP.route('/draw')
def draw():
	""" Wrapper to parse & draw tree(s). """
	if len(request.args['tree']) > LIMIT:
		return 'Too much data. Limit: %d bytes' % LIMIT
	dts = [DrawTree(tree, sent, abbr='abbr' in request.args)
				for tree, sent in incrementaltreereader(
					request.args['tree'].splitlines())]
	return drawtrees(request.args, dts)


def drawtrees(form, dts):
	""" Draw trees in the requested format. """
	os.chdir('/tmp')
	if form.get('type', 'matrix') == 'matrix':
		latexcode = ''.join((PREAMBLE % (
			'\\usepackage{helvet,tikz}\n'
			'\\usetikzlibrary{matrix,positioning}'),
			'\n\n'.join(
				dt.tikzmatrix(
						leafcolor=('blue' if form.get('color', '') else ''),
						nodecolor=('red' if form.get('color', '') else ''))
				for dt in dts),
				POSTAMBLE))
	elif form.get('type', None) == 'qtree':
		result = [PREAMBLE % r'\usepackage{helvet,tikz,tikz-qtree}']
		for dt in dts:
			for pos, word in zip(
					dt.tree.subtrees(lambda n: n and isinstance(n[0], int)),
					dt.sent):
				pos[0] = word
			result.append('\n\n'.join(dt.tree.pprint_latex_qtree()
					for dt in dts))
		result.append(POSTAMBLE)
		latexcode = ''.join(result)
	else:
		latexcode = ''.join((PREAMBLE % (
			'\\usepackage{helvet,tikz}\n'
			'\\usetikzlibrary{positioning}'),
			'\n\n'.join(
				dt.tikznode(
						leafcolor='blue' if form.get('color', '') else '',
						nodecolor='red' if form.get('color', '') else '')
				for dt in dts),
				POSTAMBLE))
	if form.get('output', 'text') == 'latex':
		return Response(latexcode, mimetype='text/plain')
	elif form.get('output', 'text') == 'svg':
		if len(dts) == 1:
			return Response(dts[0].svg().encode('utf-8'),
					mimetype='image/svg+xml')
		else:
			result = [('<!doctype html>\n<html>\n<head>\n'
				'\t<meta http-equiv="Content-Type" '
				'content="text/html; charset=UTF-8">\n</head>\n<body>')]
			for dt in dts:
				result.append(
						'<div>\n%s\n</div>\n\n' % dt.svg().encode('utf-8'))
			result.append('</body></html>')
			return Response('\n'.join(result), mimetype='text/html')
	elif form.get('output', 'text') == 'text':
		html = form.get('color', False)
		ascii = not form.get('unicode', 0)
		result = []
		if html:
			mimetype = 'text/html'
			result.append('<!doctype html>\n<html>\n<head>\n'
				'\t<meta http-equiv="Content-Type" '
				'content="text/html; charset=UTF-8">\n</head>\n<body>\n<pre>')
		else:
			mimetype = 'text/plain'
		for dt in dts:
			result.append(
					dt.text(unicodelines=not ascii, html=html).encode('utf-8'))
		if html:
			result.append('</pre></body></html>')
		return Response('\n'.join(result), mimetype=mimetype)
	else:
		with open('/tmp/dtree.tex', 'w') as tex:
			tex.write(latexcode)
		proc = Popen('/usr/bin/pdflatex -halt-on-error  /tmp/dtree.tex'.split(),
				stdin=None, stdout=PIPE, stderr=PIPE, shell=False)
		proc.wait()  # pylint: disable=E1101
		if form.get('output', 'text') == 'pdf':
			try:
				return send_from_directory('/tmp', 'dtree.pdf',
						mimetype='application/pdf')
			except IOError:
				pass
			for ext in ('aux', 'log', 'pdf'):
				os.remove('/tmp/dtree.' + ext)
		if form.get('output', 'text') == 'png':
			proc = Popen('/usr/bin/convert -density 125 /tmp/dtree.pdf \
					/tmp/dtree.png'.split(),  # -trim
					stdin=None, stdout=PIPE, stderr=PIPE, shell=False)
			proc.wait()  # pylint: disable=E1101
			#set Expires one day ahead (according to server time)
			#req.headers_out['Expires'] = (
			#	datetime.utcnow() + timedelta( 1, 0 )
			#	).strftime('%a, %d %b %Y %H:%M:%S UTC')
			#req.headers_out['Cache-Control'] = 'max-age=604800, public'
			try:
				return send_from_directory('/tmp', 'dtree.png',
						mimetype='image/png')
			except IOError:
				pass
			for filename in ('/tmp/dtree.aux', '/tmp/dtree.log',
					'/tmp/dtree.tex', '/tmp/dtree.pdf', '/tmp/dtree.png'):
				try:
					os.remove(filename)
				except OSError:
					break


if __name__ == '__main__':
	APP.run(debug=True, host='0.0.0.0')



""" Data types for chart items, edges, &c. """
from math import exp, log, fsum
from libc.math cimport log, exp
from discodop.tree import Tree

DEF SLOTS = 3
maxbitveclen = SLOTS * sizeof(ULong) * 8


cdef class LexicalRule:
	""" A weighted rule of the form 'non-terminal --> word'. """
	def __init__(self, lhs, word, prob):
		self.lhs = lhs
		self.word = word
		self.prob = prob

	def __repr__(self):
		return "%s%r" % (self.__class__.__name__,
				(self.lhs, self.word, self.prob))


cdef class SmallChartItem:
	""" Item with word sized bitvector """
	def __init__(SmallChartItem self, label, vec):
		self.label = label
		self.vec = vec

	def __hash__(SmallChartItem self):
		""" juxtapose bits of label and vec, rotating vec if > 33 words:
		64              32            0
		|               ..........label
		|vec[0] 1st half
		|               vec[0] 2nd half
		------------------------------- XOR """
		return (self.label ^ (self.vec << (sizeof(self.vec) / 2 - 1))
				^ (self.vec >> (sizeof(self.vec) / 2 - 1)))

	def __richcmp__(SmallChartItem self, SmallChartItem ob, int op):
		if op == 2:
			return self.label == ob.label and self.vec == ob.vec
		elif op == 3:
			return self.label != ob.label or self.vec != ob.vec
		elif op == 5:
			return self.label >= ob.label or self.vec >= ob.vec
		elif op == 1:
			return self.label <= ob.label or self.vec <= ob.vec
		elif op == 0:
			return self.label < ob.label or self.vec < ob.vec
		elif op == 4:
			return self.label > ob.label or self.vec > ob.vec

	def __nonzero__(SmallChartItem self):
		return self.label != 0 and self.vec != 0

	def __repr__(self):
		return "%s(%d, %s)" % (self.__class__.__name__,
				self.label, self.binrepr())

	def lexidx(self):
		assert self.label == 0
		return self.vec

	def copy(SmallChartItem self):
		return SmallChartItem(self.label, self.vec)

	def binrepr(SmallChartItem self, int lensent=0):
		return bin(self.vec)[2:].zfill(lensent)[::-1]


cdef class FatChartItem:
	""" Item with fixed-with bitvector. """
	def __hash__(self):
		cdef long n, _hash
		""" juxtapose bits of label and vec:
		64              32            0
		|               ..........label
		|vec[0] 1st half
		|               vec[0] 2nd half
		|........ rest of vec .........
		------------------------------- XOR """
		_hash = (self.label ^ self.vec[0] << (8 * sizeof(self.vec[0]) / 2 - 1)
				^ self.vec[0] >> (8 * sizeof(self.vec[0]) / 2 - 1))
		# add remaining bits
		for n in range(sizeof(self.vec[0]), sizeof(self.vec)):
			_hash *= 33 ^ (<UChar *>self.vec)[n]
		return _hash

	def __richcmp__(FatChartItem self, FatChartItem ob, int op):
		cdef int cmp = memcmp(<UChar *>self.vec, <UChar *>ob.vec,
			sizeof(self.vec))
		cdef bint labelmatch = self.label == ob.label
		if op == 2:
			return labelmatch and cmp == 0
		elif op == 3:
			return not labelmatch or cmp != 0
		elif op == 5:
			return self.label >= ob.label or (labelmatch and cmp >= 0)
		elif op == 1:
			return self.label <= ob.label or (labelmatch and cmp <= 0)
		elif op == 0:
			return self.label < ob.label or (labelmatch and cmp < 0)
		elif op == 4:
			return self.label > ob.label or (labelmatch and cmp > 0)

	def __nonzero__(self):
		cdef int n
		if self.label:
			for n in range(SLOTS):
				if self.vec[n]:
					return True
		return False

	def __repr__(self):
		return "%s(%d, %s)" % (self.__class__.__name__,
			self.label, self.binrepr())

	def lexidx(self):
		assert self.label == 0
		return self.vec[0]

	def copy(FatChartItem self):
		cdef FatChartItem a = FatChartItem(self.label)
		for n in range(SLOTS):
			a.vec[n] = self.vec[n]
		return a

	def binrepr(FatChartItem self, lensent=0):
		cdef int m, n = SLOTS - 1
		cdef str result
		while n and self.vec[n] == 0:
			n -= 1
		result = bin(self.vec[n])
		for m in range(n - 1, -1, -1):
			result += bin(self.vec[m])[2:].zfill(BITSIZE)
		return result.zfill(lensent)[::-1]


cdef class CFGChartItem:
	""" Item for CFG parsing; span is denoted with start and end indices. """
	def __init__(self, label, start, end):
		self.label = label
		self.start = start
		self.end = end

	def __hash__(self):
		""" juxtapose bits of label and indices of span:
		|....end...start...label
		64    40      32       0 """
		return (self.label ^ <ULong>self.start << (8 * sizeof(long) / 2)
				^ <ULong>self.end << (8 * sizeof(long) / 2 + 8))

	def __richcmp__(CFGChartItem self, CFGChartItem ob, int op):
		cdef bint labelmatch = self.label == ob.label
		cdef bint startmatch = self.start == ob.start
		if op == 2:
			return labelmatch and startmatch and self.end == ob.end
		elif op == 3:
			return not labelmatch or not startmatch or self.end != ob.end
		elif op == 5:
			return self.label >= ob.label or (labelmatch and (self.start
					>= ob.start or startmatch and self.end >= ob.end))
		elif op == 1:
			return self.label <= ob.label or (labelmatch and (self.start
					<= ob.start or (startmatch and self.end <= ob.end)))
		elif op == 0:
			return self.label < ob.label or (labelmatch and (self.start
					< ob.start or (startmatch and self.end < ob.end)))
		elif op == 4:
			return self.label > ob.label or (labelmatch and (self.start
					> ob.start or (startmatch and self.end > ob.end)))

	def __nonzero__(self):
		return self.label and self.end

	def __repr__(self):
		return "%s(%d, %d, %d)" % (self.__class__.__name__,
				self.label, self.start, self.end)

	def lexidx(self):
		assert self.label == 0
		return self.start

	def copy(CFGChartItem self):
		return new_CFGChartItem(self.label, self.start, self.end)


cdef SmallChartItem CFGtoSmallChartItem(UInt label, UChar start, UChar end):
	return new_ChartItem(label, (1ULL << end) - (1ULL << start))


cdef FatChartItem CFGtoFatChartItem(UInt label, UChar start, UChar end):
	cdef FatChartItem fci = new_FatChartItem(label)
	if BITSLOT(start) == BITSLOT(end):
		fci.vec[BITSLOT(start)] = (1ULL << end) - (1ULL << start)
	else:
		fci.vec[BITSLOT(start)] = ~0UL << (start % BITSIZE)
		for n in range(BITSLOT(start) + 1, BITSLOT(end)):
			fci.vec[n] = ~0UL
		fci.vec[BITSLOT(end)] = BITMASK(end) - 1
	return fci


cdef class LCFRSEdge:
	""" NB: hash / (in)equality considers all elements except inside score,
	order is determined by inside score only. """
	def __hash__(LCFRSEdge self):
		cdef long _hash = 0x345678UL
		# this condition could be avoided by using a dedicated sentinel Rule
		if self.rule is not NULL:
			_hash = (1000003UL * _hash) ^ <long>self.rule.no
		# we only look at the left item, because this edge will only be
		# compared to other edges for the same parent item
		# FIXME: we cannot compute hash directly here, because
		# left can be of different subtypes.
		_hash = (1000003UL * _hash) ^ <long>self.left.__hash__()
		return _hash

	def __richcmp__(LCFRSEdge self, LCFRSEdge ob, int op):
		if op == 0:
			return self.score < ob.score
		elif op == 1:
			return self.score <= ob.score
		elif op == 2 or op == 3:
			# right matches iff left matches, so skip that check
			return (op == 2) == (self.rule is ob.rule and self.left == ob.left)
		elif op == 4:
			return self.score > ob.score
		elif op == 5:
			return self.score >= ob.score
		elif op == 1:
			return self.score <= ob.score
		elif op == 0:
			return self.score < ob.score

	def __repr__(self):
		return "%s(%g, %g, Rule(%g, 0x%x, 0x%x, %d, %d, %d, %d), %r, %r)" % (
				self.__class__.__name__, self.score, self.inside,
				self.rule.prob, self.rule.args, self.rule.lengths,
				self.rule.lhs, self.rule.rhs1, self.rule.rhs2, self.rule.no,
				self.left, self.right)

	def copy(self):
		return new_LCFRSEdge(self.score, self.inside, self.rule,
				self.left.copy(), self.right.copy())


cdef class CFGEdge:
	""" NB: hash / (in)equality considers all elements except inside score,
	order is determined by inside score only. """
	def __hash__(CFGEdge self):
		cdef long _hash = 0x345678UL
		_hash = (1000003UL * _hash) ^ <long>self.rule
		_hash = (1000003UL * _hash) ^ <long>self.mid
		return _hash

	def __richcmp__(CFGEdge self, CFGEdge ob, int op):
		if op == 0:
			return self.inside < ob.inside
		elif op == 1:
			return self.inside <= ob.inside
		elif op == 2 or op == 3:
			return (op == 2) == (self.rule is ob.rule and self.mid == ob.mid)
		elif op == 4:
			return self.inside > ob.inside
		elif op == 5:
			return self.inside >= ob.inside
		elif op == 1:
			return self.inside <= ob.inside
		elif op == 0:
			return self.inside < ob.inside

	def __repr__(self):
		return "%s(%g, Rule(%g, 0x%x, 0x%x, %d, %d, %d, %d), %r)" % (
			self.__class__.__name__, self.inside, self.rule.prob,
			self.rule.args, self.rule.lengths, self.rule.lhs, self.rule.rhs1,
			self.rule.rhs2, self.rule.no, self.mid)


cdef class RankedEdge:
	""" An edge, including the ChartItem to which it points, along with
	ranks for its children, to denote a k-best derivation. """
	def __cinit__(self, ChartItem head, LCFRSEdge edge, int j1, int j2):
		self.head = head
		self.edge = edge
		self.left = j1
		self.right = j2

	def __hash__(self):
		cdef long _hash = 0x345678UL
		_hash = (1000003UL * _hash) ^ hash(self.head)
		_hash = (1000003UL * _hash) ^ hash(self.edge)
		_hash = (1000003UL * _hash) ^ self.left
		_hash = (1000003UL * _hash) ^ self.right
		return _hash

	def __richcmp__(RankedEdge self, RankedEdge ob, int op):
		if op == 2 or op == 3:
			return (op == 2) == (self.left == ob.left and self.right ==
					ob.right and self.head == ob.head and self.edge == ob.edge)
		return NotImplemented

	def __repr__(self):
		return "%s(%r, %r, %d, %d)" % (self.__class__.__name__,
			self.head, self.edge, self.left, self.right)


cdef class RankedCFGEdge:
	""" An edge, including the ChartItem to which it points, along with
	ranks for its children, to denote a k-best derivation. """
	def __cinit__(self, UInt label, UChar start, UChar end, Edge edge,
			int j1, int j2):
		self.label = label
		self.start = start
		self.end = end
		self.edge = edge
		self.left = j1
		self.right = j2

	def __hash__(self):
		cdef long _hash = 0x345678UL
		_hash = (1000003UL * _hash) ^ hash(self.edge)
		_hash = (1000003UL * _hash) ^ self.label
		_hash = (1000003UL * _hash) ^ self.start
		_hash = (1000003UL * _hash) ^ self.end
		_hash = (1000003UL * _hash) ^ self.left
		_hash = (1000003UL * _hash) ^ self.right
		return _hash

	def __richcmp__(RankedCFGEdge self, RankedCFGEdge ob, int op):
		if op == 2 or op == 3:
			return (op == 2) == (self.left == ob.left and self.right ==
					ob.right and self.label == ob.label and self.start ==
					ob.start and self.end == ob.end and self.edge == ob.edge)
		return NotImplemented

	def __repr__(self):
		return "%s(%r, %r, %r, %r, %d, %d)" % (self.__class__.__name__,
			self.label, self.start, self.end, self.edge, self.left, self.right)


cdef class Ctrees:
	"""
	Auxiliary class to be able to pass around collections of NodeArrays
	in Python.

	When trees is given, prods should be given as well.
	When trees is not given, the alloc() method should be called and
	trees added one by one using the add() or addnodes() methods. """
	def __cinit__(self):
		self.trees = self.nodes = NULL

	def __init__(self, list trees=None, dict prods=None):
		self.len = self.max = 0
		self.numnodes = self.maxnodes = self.nodesleft = 0
		if trees is not None:
			assert prods is not None
			self.alloc(len(trees), sum(map(len, trees)))
			for tree in trees:
				self.add(tree, prods)

	cpdef alloc(self, int numtrees, long numnodes):
		""" Initialize an array of trees of nodes structs. """
		self.max = numtrees
		self.trees = <NodeArray *>malloc(numtrees * sizeof(NodeArray))
		self.nodes = <Node *>malloc(numnodes * sizeof(Node))
		assert self.trees is not NULL and self.nodes is not NULL
		self.nodesleft = numnodes

	cdef realloc(self, int len):
		""" Increase size of array (handy with incremental binarization) """
		self.nodesleft += len
		#estimate how many new nodes will be needed
		self.nodesleft += (self.max - self.len) * (self.numnodes / self.len)
		self.nodes = <Node *>realloc(self.nodes,
				(self.numnodes + self.nodesleft) * sizeof(Node))
		assert self.nodes is not NULL

	cpdef add(self, list tree, dict prods):
		""" Trees can be incrementally added to the node array; useful
		when dealing with large numbers of NLTK trees (say 100,000). """
		assert self.len < self.max, ("either no space left (len >= max) or "
			"alloc() has not been called (max=0). max = %d" % self.max)
		if self.nodesleft < len(tree):
			self.realloc(len(tree))
		self.trees[self.len].len = len(tree)
		self.trees[self.len].offset = self.numnodes
		copynodes(tree, prods, &self.nodes[self.numnodes])
		self.trees[self.len].root = tree[0].rootidx
		self.len += 1
		self.nodesleft -= len(tree)
		self.numnodes += len(tree)
		self.maxnodes = max(self.maxnodes, len(tree))

	cdef addnodes(self, Node *source, int cnt, int root):
		""" Trees can be incrementally added to the node array; this version
		copies a tree that has already been converted to an array of nodes. """
		cdef dict prodsintree, sortidx
		cdef int n, m
		cdef Node *dest
		assert self.len < self.max, ("either no space left (len >= max) or "
				"alloc() has not been called (max=0).\n"
				"len = %d, max = %d" % (self.len, self.max))
		if self.nodesleft < cnt:
			self.realloc(cnt)
		prodsintree = {n: source[n].prod for n in range(cnt)}
		sortidx = {m: n for n, m in enumerate(
				sorted(range(cnt), key=prodsintree.get))}
		# copy nodes to allocated array, while translating indices
		dest = &self.nodes[self.numnodes]
		for n, m in sortidx.iteritems():
			dest[m] = source[n]
			if dest[m].left >= 0:
				dest[m].left = sortidx[source[n].left]
				if dest[m].right >= 0:
					dest[m].right = sortidx[source[n].right]
		self.trees[self.len].offset = self.numnodes
		self.trees[self.len].root = sortidx[root]
		self.trees[self.len].len = cnt
		self.len += 1
		self.nodesleft -= cnt
		self.numnodes += cnt
		if cnt > self.maxnodes:
			self.maxnodes = cnt

	def indextrees(self, dict prods):
		""" Create an index from specific productions to trees containing that
		production. Productions are represented as integer IDs, trees are given
		as sets of integer indices. """
		cdef:
			list result = [set() for _ in prods]
			NodeArray a
			Node *nodes
			int n, m
		for n in range(self.len):
			a = self.trees[n]
			nodes = &self.nodes[a.offset]
			for m in range(a.len):
				(<set>result[nodes[m].prod]).add(n)
		self.treeswithprod = result

	def __dealloc__(Ctrees self):
		if self.nodes is not NULL:
			free(self.nodes)
			self.nodes = NULL
		if self.trees is not NULL:
			free(self.trees)
			self.trees = NULL

	def __len__(self):
		return self.len


cdef inline copynodes(tree, dict prods, Node *result):
	""" Convert NLTK tree to an array of Node structs. """
	cdef int n
	for n, a in enumerate(tree):
		assert isinstance(a, Tree), (
				'Expected Tree node, got %s\n%r' % (type(a), a))
		assert 1 <= len(a) <= 2, (
				"trees must be non-empty and binarized\n%s\n%s" % (a, tree[0]))
		result[n].prod = prods[a.prod]
		if isinstance(a[0], int):  # a terminal index
			result[n].left = -a[0] - 1
		else:
			result[n].left = a[0].idx
			if len(a) == 2:
				result[n].right = a[1].idx
			else:  # unary node
				result[n].right = -1



""" Implementation of outside estimates:
- PCFG A* estimate (Klein & Manning 2003),
- PLCFRS LR context-summary estimate (Kallmeyer & Maier 2010).
The latter ported almost directly from rparse
(except for sign reversal of log probs). """

from __future__ import print_function
from math import exp
import numpy as np

from discodop.bit cimport nextset, nextunset, bitcount, bitlength, testbit, \
		testbitint
from discodop.agenda cimport Agenda, Entry
from discodop._grammar cimport Grammar, Rule, LexicalRule, UInt, ULLong
from discodop.containers cimport new_ChartItem, SmallChartItem as ChartItem

cdef extern from "math.h":
	bint isnan(double x)
	bint isfinite(double x)

INFINITY = float('infinity')

cdef class Item:
	""" Item class used in agenda for computing the  outside LR estimate. """
	cdef int state, length, lr, gaps

	def __hash__(Item self):
		return self.state * 1021 + self.length * 571 + self.lr * 311 + self.gaps

	def __richcmp__(Item self, Item other, op):
		if op == 2:
			return (self.state == other.state and self.length == other.length
				and self.lr == other.lr and self.gaps == other.gaps)
		a = (self.state, self.length, self.lr, self.gaps)
		b = (other.state, other.length, other.lr, other.gaps)
		if op == 0:
			return a < b
		elif op == 1:
			return a <= b
		elif op == 3:
			return a != b
		elif op == 4:
			return a > b
		elif op == 5:
			return a >= b

	def __repr__(Item self):
		return "%s(%4d, %2d, %2d, %2d)" % (self.__class__.__name__,
			self.state, self.length, self.lr, self.gaps)


cdef inline Item new_Item(int state, int length, int lr, int gaps):
	cdef Item item = Item.__new__(Item)
	item.state = state
	item.length = length
	item.lr = lr
	item.gaps = gaps
	return item


cdef inline double getoutside(double [:, :, :, :] outside,
		UInt maxlen, UInt slen, UInt label, ULLong vec):
	""" Query for outside estimate. NB: if this would be used, it should be in
	a .pxd with `inline`. However, passing the numpy array may be slow. """
	cdef UInt length = bitcount(vec)
	cdef UInt left = nextset(vec, 0)
	cdef UInt gaps = bitlength(vec) - length - left
	cdef UInt right = slen - length - left - gaps
	cdef UInt lr = left + right
	if slen > maxlen or length + lr + gaps > maxlen:
		return 0.0
	return outside[label, length, lr, gaps]


def simpleinside(Grammar grammar, UInt maxlen, double [:, :] insidescores):
	""" Compute simple inside estimate in bottom-up fashion.
	Here vec is actually the length (number of terminals in the yield of
	the constituent)
	insidescores is a 2-dimensional matrix initialized with NaN to indicate
	values that have yet to be computed. """
	cdef double x
	cdef Agenda agenda = Agenda()
	cdef ChartItem I
	cdef Entry entry
	cdef LexicalRule lexrule
	cdef Rule rule
	cdef size_t i
	cdef ULLong vec

	for i in grammar.lexicalbylhs:
		agenda[new_ChartItem(i, 1)] = min([lexrule.prob
			for lexrule in grammar.lexicalbylhs[i].values()])

	while agenda.length:
		entry = agenda.popentry()
		I = entry.key
		x = entry.value

		# This comparison catches the case when insidescores has a higher
		# value than x, but also when it is NaN, because all comparisons
		# against NaN are false.
		#if not x >= insidescores[I.label, I.vec]:
		# Mory explicitly:
		if (isnan(insidescores[I.label, I.vec])
			or x < insidescores[I.label, I.vec]):
			insidescores[I.label, I.vec] = x

		for i in range(grammar.nonterminals):
			rule = grammar.unary[I.label][i]
			if rule.rhs1 != I.label:
				break
			elif isnan(insidescores[rule.lhs, I.vec]):
				agenda.setifbetter(
						new_ChartItem(rule.lhs, I.vec), rule.prob + x)

		for i in range(grammar.nonterminals):
			rule = grammar.lbinary[I.label][i]
			if rule.rhs1 != I.label:
				break
			for vec in range(1, maxlen - I.vec + 1):
				if (isfinite(insidescores[rule.rhs2, vec])
					and isnan(insidescores[rule.lhs, I.vec + vec])):
					agenda.setifbetter(new_ChartItem(rule.lhs, I.vec + vec),
						rule.prob + x + insidescores[rule.rhs2, vec])

		for i in range(grammar.nonterminals):
			rule = grammar.rbinary[I.label][i]
			if rule.rhs2 != I.label:
				break
			for vec in range(1, maxlen - I.vec + 1):
				if (isfinite(insidescores[rule.rhs1, vec])
					and isnan(insidescores[rule.lhs, vec + I.vec])):
					agenda.setifbetter(new_ChartItem(rule.lhs, vec + I.vec),
						rule.prob + insidescores[rule.rhs1, vec] + x)

	# anything not reached so far is still NaN and gets probability zero:
	insidescores.base[np.isnan(insidescores.base)] = np.inf


def inside(Grammar grammar, UInt maxlen, dict insidescores):
	""" Compute inside estimate in bottom-up fashion, with
	full bit vectors (not used). """
	cdef ChartItem I
	cdef Entry entry
	cdef LexicalRule lexrule
	cdef size_t i
	agenda = Agenda()

	for i in grammar.lexicalbylhs:
		agenda[new_ChartItem(i, 1)] = min([lexrule.prob
			for lexrule in grammar.lexicalbylhs[i].values()])

	while agenda.length:
		entry = agenda.popentry()
		I = entry.key
		x = entry.value
		if I.label not in insidescores:
			insidescores[I.label] = {}
		if x < insidescores[I.label].get(I.vec, INFINITY):
			insidescores[I.label][I.vec] = x

		for i in range(grammar.nonterminals):
			rule = grammar.unary[I.label][i]
			if rule.rhs1 != I.label:
				break
			elif (rule.lhs not in insidescores
				or I.vec not in insidescores[rule.lhs]):
				agenda.setifbetter(
					new_ChartItem(rule.lhs, I.vec), rule.prob + x)

		for i in range(grammar.nonterminals):
			rule = grammar.lbinary[I.label][i]
			if rule.rhs1 != I.label:
				break
			elif rule.rhs2 not in insidescores:
				continue
			for vec in insidescores[rule.rhs2]:
				left = insideconcat(I.vec, vec, rule, grammar, maxlen)
				if left and (rule.lhs not in insidescores
					or left not in insidescores[rule.lhs]):
					agenda.setifbetter(new_ChartItem(rule.lhs, left),
						rule.prob + x + insidescores[rule.rhs2][vec])

		for i in range(grammar.nonterminals):
			rule = grammar.rbinary[I.label][i]
			if rule.rhs2 != I.label:
				break
			elif rule.rhs1 not in insidescores:
				continue
			for vec in insidescores[rule.rhs1]:
				right = insideconcat(vec, I.vec, rule, grammar, maxlen)
				if right and (rule.lhs not in insidescores
					or right not in insidescores[rule.lhs]):
					agenda.setifbetter(new_ChartItem(rule.lhs, right),
						rule.prob + insidescores[rule.rhs1][vec] + x)

	return insidescores


cdef inline ULLong insideconcat(ULLong a, ULLong b, Rule rule, Grammar grammar,
		UInt maxlen):
	if grammar.fanout[rule.lhs] + bitcount(a) + bitcount(b) > maxlen + 1:
		return 0
	result = resultpos = l = r = 0
	for x in range(bitlength(rule.lengths)):
		if testbitint(rule.args, x) == 0:
			subarg = nextunset(a, l) - l
			result |= (1 << subarg) - 1 << resultpos
			resultpos += subarg
			l = subarg + 1
		else:
			subarg = nextunset(b, r) - r
			result |= (1 << subarg) - 1 << resultpos
			resultpos += subarg
			r = subarg + 1
		if testbitint(rule.lengths, x):
			resultpos += 1
			result &= ~(1 << resultpos)
	return result


def outsidelr(Grammar grammar, double [:, :] insidescores,
		UInt maxlen, UInt goal, double [:, :, :, :] outside):
	""" Compute the outside SX simple LR estimate in top down fashion. """
	cdef Agenda agenda = Agenda()
	cdef Entry entry
	cdef Item I
	cdef Rule rule
	cdef LexicalRule lexrule
	cdef double x, insidescore, current, score
	cdef int m, n, totlen, addgaps, addright, leftfanout, rightfanout
	cdef int lenA, lenB, lr, ga
	cdef size_t i
	cdef bint stopaddleft, stopaddright

	for n in range(1, maxlen + 1):
		agenda[new_Item(goal, n, 0, 0)] = 0.0
		outside[goal, n, 0, 0] = 0.0
	print("initialized")

	while agenda.length:
		entry = agenda.popentry()
		I = entry.key
		x = entry.value
		if agenda.length % 1000 == 0:
			print("agenda size: %dk top: %r, %g %s" % (
				agenda.length / 1000, I, exp(-x),
				grammar.tolabel[I.state].decode('ascii')))
		totlen = I.length + I.lr + I.gaps
		i = 0
		rule = grammar.bylhs[I.state][i]
		while rule.lhs == I.state:
			# X -> A
			if rule.rhs2 == 0:
				score = rule.prob + x
				if score < outside[rule.rhs1, I.length, I.lr, I.gaps]:
					agenda.setitem(
						new_Item(rule.rhs1, I.length, I.lr, I.gaps), score)
					outside[rule.rhs1, I.length, I.lr, I.gaps] = score
				i += 1
				rule = grammar.bylhs[I.state][i]
				continue
			# X -> A B
			addgaps = addright = 0
			stopaddright = False
			for n in range(bitlength(rule.lengths) - 1, -1, -1):
				if testbitint(rule.args, n):
					if stopaddright:
						addgaps += 1
					else:
						addright += 1
				elif not stopaddright:
					stopaddright = True

			leftfanout = grammar.fanout[rule.rhs1]
			rightfanout = grammar.fanout[rule.rhs2]

			# binary-left (A is left)
			for lenA in range(leftfanout, I.length - rightfanout + 1):
				lenB = I.length - lenA
				insidescore = insidescores[rule.rhs2, lenB]
				for lr in range(I.lr, I.lr + lenB + 2):  # FIXME: why 2?
					if addright == 0 and lr != I.lr:
						continue
					for ga in range(leftfanout - 1, totlen + 1):
						if (lenA + lr + ga == I.length + I.lr + I.gaps
							and ga >= addgaps):
							score = rule.prob + x + insidescore
							current = outside[rule.rhs1, lenA, lr, ga]
							if score < current:
								agenda.setitem(
									new_Item(rule.rhs1, lenA, lr, ga), score)
								outside[rule.rhs1, lenA, lr, ga] = score

			# X -> B A
			addgaps = addright = 0
			stopaddleft = False
			for n in range(bitlength(rule.lengths)):
				if not stopaddleft and testbitint(rule.args, n):
					stopaddleft = True
				if not testbitint(rule.args, n):
					if stopaddleft:
						addgaps += 1

			stopaddright = False
			for n in range(bitlength(rule.lengths) - 1, -1, -1):
				if not stopaddright and testbitint(rule.args, n):
					stopaddright = True
				if not testbitint(rule.args, n) and not stopaddright:
					addright += 1
			addgaps -= addright

			# binary-right (A is right)
			for lenA in range(rightfanout, I.length - leftfanout + 1):
				lenB = I.length - lenA
				insidescore = insidescores[rule.rhs1, lenB]
				for lr in range(I.lr, I.lr + lenB + 2):  # FIXME: why 2?
					for ga in range(rightfanout - 1, totlen + 1):
						if (lenA + lr + ga == I.length + I.lr + I.gaps
							and ga >= addgaps):
							score = rule.prob + insidescore + x
							current = outside[rule.rhs2, lenA, lr, ga]
							if score < current:
								agenda.setitem(
									new_Item(rule.rhs2, lenA, lr, ga), score)
								outside[rule.rhs2, lenA, lr, ga] = score
			i += 1
			rule = grammar.bylhs[I.state][i]
		#end while rule.lhs == I.state:
	#end while agenda.length:


def getestimates(Grammar grammar, UInt maxlen, UInt goal):
	""" Compute table of outside SX simple LR estimates for a PLCFRS. """
	print("allocating outside matrix:",
		(8 * grammar.nonterminals * (maxlen + 1) * (maxlen + 1)
			* (maxlen + 1) / 1024 ** 2), 'MB')
	insidescores = np.empty((grammar.nonterminals, (maxlen + 1)), dtype='d')
	outside = np.empty((grammar.nonterminals, ) + 3 * (maxlen + 1, ), dtype='d')
	insidescores[...] = np.NAN
	outside[...] = np.inf
	print("getting inside estimates")
	simpleinside(grammar, maxlen, insidescores)
	print("getting outside estimates")
	outsidelr(grammar, insidescores, maxlen, goal, outside)
	return outside


cdef inline double getpcfgoutside(dict outsidescores,
		UInt maxlen, UInt slen, UInt label, ULLong vec):
	""" Query for a PCFG A* estimate. For documentation purposes. """
	cdef UInt length = bitcount(vec)
	cdef UInt left = nextset(vec, 0)
	cdef UInt right = slen - length - left
	cdef UInt lr = left + right
	if slen > maxlen or length + left + right > maxlen:
		return 0.0
	return outsidescores[label, left, right]


cpdef getpcfgestimates(Grammar grammar, UInt maxlen, UInt goal,
		bint debug=False):
	""" Compute table of outside SX estimates for a PCFG. """
	insidescores = pcfginsidesx(grammar, maxlen)
	outside = pcfgoutsidesx(grammar, insidescores, goal, maxlen)
	if debug:
		print('inside:')
		for span in range(1, maxlen + 1):
			for k, v in sorted(insidescores[span].items()):
				if v < INFINITY:
					print("%s[%d] %g" % (grammar.tolabel[k].decode('ascii'),
						span, exp(-v)))
		print('infinite:', end='')
		for span in range(1, maxlen + 1):
			for k, v in sorted(insidescores[span].items()):
				if v == INFINITY:
					print("%s[%d]" % (grammar.tolabel[k].decode('ascii'),
							span), end='')
		print('\n\noutside:')
		for lspan in range(maxlen + 1):
			for rspan in range(maxlen - lspan + 1):
				for lhs in range(grammar.nonterminals):
					if outside[lhs, lspan, rspan] < INFINITY:
						print("%s[%d-%d] %g" % (
								grammar.tolabel[lhs].decode('ascii'), lspan,
								rspan, exp(-outside[lhs, lspan, rspan])))
	return outside


cdef pcfginsidesx(Grammar grammar, UInt maxlen):
	""" insideSX estimate for a PCFG using agenda. Adapted from:
	Klein & Manning (2003), A* parsing: Fast Exact Viterbi Parse Selection. """
	cdef size_t n, split
	cdef ULLong vec
	cdef ChartItem I
	cdef Entry entry
	cdef Rule rule
	cdef LexicalRule lexrule
	cdef Agenda agenda = Agenda()
	cdef double x
	cdef list insidescores = [{} for n in range(maxlen + 1)]
	for n in grammar.lexicalbylhs:
		x = min([lexrule.prob for lexrule in grammar.lexicalbylhs[n].values()])
		agenda[new_ChartItem(n, 1)] = x
	while agenda.length:
		entry = agenda.popentry()
		I = entry.key
		x = entry.value
		if (I.label not in insidescores[I.vec]
				or x < insidescores[I.vec][I.label]):
			insidescores[I.vec][I.label] = x

		for i in range(grammar.nonterminals):
			rule = grammar.unary[I.label][i]
			if rule.rhs1 != I.label:
				break
			elif rule.lhs not in insidescores[I.vec]:
				agenda.setifbetter(
						new_ChartItem(rule.lhs, I.vec), rule.prob + x)

		for i in range(grammar.nonterminals):
			rule = grammar.lbinary[I.label][i]
			if rule.rhs1 != I.label:
				break
			for vec in range(1, maxlen - I.vec + 1):
				if (rule.rhs2 in insidescores[vec]
					and rule.lhs not in insidescores[I.vec + vec]):
					agenda.setifbetter(new_ChartItem(rule.lhs, I.vec + vec),
						rule.prob + x + insidescores[vec][rule.rhs2])

		for i in range(grammar.nonterminals):
			rule = grammar.rbinary[I.label][i]
			if rule.rhs2 != I.label:
				break
			for vec in range(1, maxlen - I.vec + 1):
				if (rule.rhs1 in insidescores[vec]
					and rule.lhs not in insidescores[vec + I.vec]):
					agenda.setifbetter(new_ChartItem(rule.lhs, vec + I.vec),
						rule.prob + insidescores[vec][rule.rhs1] + x)
	return insidescores


cdef pcfgoutsidesx(Grammar grammar, list insidescores, UInt goal, UInt maxlen):
	""" outsideSX estimate for a PCFG, agenda-based version. """
	cdef Agenda agenda = Agenda()
	cdef Entry entry
	cdef tuple I
	cdef Rule rule
	cdef LexicalRule lexrule
	cdef double x, insidescore, current, score
	cdef int m, n, state, left, right
	cdef size_t i, sibsize
	cdef double [:, :, :, :] outside = np.empty(
			(grammar.nonterminals, maxlen + 1, maxlen + 1, 1), dtype='d')
	outside[...] = np.inf

	agenda[goal, 0, 0] = outside[goal, 0, 0, 0] = 0.0
	while agenda.length:
		entry = agenda.popentry()
		I = entry.key
		x = entry.value
		state, left, right = I
		if agenda.length % 1000 == 0:
			print("agenda size: %dk top: %r, %g %s" % (
				agenda.length / 1000, I, exp(-x),
				grammar.tolabel[state].decode('ascii')))
		i = 0
		rule = grammar.bylhs[state][i]
		while rule.lhs == state:
			# X -> A
			if rule.rhs2 == 0:
				score = rule.prob + x
				if score < outside[rule.rhs1, left, right, 0]:
					agenda.setitem((rule.rhs1, left, right), score)
					outside[rule.rhs1, left, right, 0] = score
				i += 1
				rule = grammar.bylhs[state][i]
				continue

			# item is on the left: X -> A B.
			for sibsize in range(1, maxlen - left - right):
				insidescore = insidescores[sibsize].get(rule.rhs2, INFINITY)
				score = rule.prob + x + insidescore
				current = outside[rule.rhs1, left, right + sibsize, 0]
				if score < current:
					agenda.setitem((rule.rhs1, left, right + sibsize), score)
					outside[rule.rhs1, left, right + sibsize, 0] = score

			# item is on the right: X -> B A
			for sibsize in range(1, maxlen - left - right):
				insidescore = insidescores[sibsize].get(rule.rhs1, INFINITY)
				score = rule.prob + insidescore + x
				current = outside[rule.rhs2, left + sibsize, right, 0]
				if score < current:
					agenda.setitem((rule.rhs2, left + sibsize, right), score)
					outside[rule.rhs2, left + sibsize, right, 0] = score

			i += 1
			rule = grammar.bylhs[state][i]
		#end while rule.lhs == state:
	#end while agenda.length:
	return outside.base


cpdef getpcfgestimatesrec(Grammar grammar, UInt maxlen, UInt goal,
	bint debug=False):
	insidescores = [{} for _ in range(maxlen + 1)]
	outsidescores = {}
	for span in range(1, maxlen + 1):
		insidescores[span][goal] = pcfginsidesxrec(
			grammar, insidescores, goal, span)
	for lspan in range(maxlen + 1):
		for rspan in range(maxlen - lspan + 1):
			for lhs in grammar.lexicalbylhs:
				if (lhs, lspan, rspan) in outsidescores:
					continue
				outsidescores[lhs, lspan, rspan] = pcfgoutsidesxrec(grammar,
						insidescores, outsidescores, goal, lhs, lspan, rspan)
	if debug:
		print('inside:')
		for span in range(1, maxlen + 1):
			for k, v in sorted(insidescores[span].items()):
				if v < INFINITY:
					print("%s[%d] %g" % (
							grammar.tolabel[k].decode('ascii'), span, exp(-v)))
		print('infinite:', end='')
		for span in range(1, maxlen + 1):
			for k, v in sorted(insidescores[span].items()):
				if v == INFINITY:
					print("%s[%d]" % (
							grammar.tolabel[k].decode('ascii'), span), end='')

		print('\n\noutside:')
		for k, v in sorted(outsidescores.items()):
			if v < INFINITY:
				print("%s[%d-%d] %g" % (
						grammar.tolabel[k[0]].decode('ascii'), k[1], k[2],
						exp(-v)))
		print('infinite:', end='')
		for k, v in sorted(outsidescores.items()):
			if v == INFINITY:
				print("%s[%d-%d]" % (
						grammar.tolabel[k[0]].decode('ascii'), k[1], k[2]),
						end='')
	outside = np.empty((grammar.nonterminals, maxlen + 1, maxlen + 1, 1),
			dtype='d')
	outside[...] = np.inf
	#convert sparse dictionary to dense numpy array
	for (state, lspan, rspan), prob in outsidescores.items():
		outside[state, lspan, rspan, 0] = prob
	return outside


cdef pcfginsidesxrec(Grammar grammar, list insidescores, UInt state, int span):
	""" insideSX estimate for a PCFG. Straight from Klein & Manning (2003),
	A* parsing: Fast Exact Viterbi Parse Selection. """
	# NB: does not deal correctly with unary rules.
	cdef size_t n, split
	cdef Rule rule
	cdef LexicalRule lexrule
	if span == 0:
		return 0 if state == 0 else INFINITY
	if span == 1 and state in grammar.lexicalbylhs:
		score = min([lexrule.prob for lexrule
				in grammar.lexicalbylhs[state].values()])
	else:
		score = INFINITY
	for split in range(1, span + 1):
		n = 0
		rule = grammar.bylhs[state][n]
		while rule.lhs == state:
			if rule.rhs1 in insidescores[split]:
				inleft = insidescores[split][rule.rhs1]
				if inleft == -1:
					n += 1
					rule = grammar.bylhs[state][n]
			else:
				insidescores[split][rule.rhs1] = -1  # mark to avoid cycles.
				inleft = pcfginsidesxrec(
					grammar, insidescores, rule.rhs1, split)
				insidescores[split][rule.rhs1] = inleft
			if rule.rhs2 == 0:
				if split == span:
					inright = 0
				else:
					n += 1
					rule = grammar.bylhs[state][n]
					continue
			elif rule.rhs2 in insidescores[span - split]:
				inright = insidescores[span - split][rule.rhs2]
			else:
				inright = pcfginsidesxrec(
					grammar, insidescores, rule.rhs2, span - split)
				insidescores[span - split][rule.rhs2] = inright
			cost = inleft + inright + rule.prob
			if cost < score:
				score = cost
			n += 1
			rule = grammar.bylhs[state][n]
	return score


cdef pcfgoutsidesxrec(Grammar grammar, list insidescores, dict outsidescores,
		UInt goal, UInt state, int lspan, int rspan):
	""" outsideSX estimate for a PCFG. """
	# NB: does not deal correctly with unary rules.
	cdef size_t n, sibsize
	cdef Rule rule
	cdef tuple item
	if lspan + rspan == 0:
		return 0 if state == goal else INFINITY
	score = INFINITY
	# unary productions: no sibling
	n = 0
	rule = grammar.unary[state][n]
	while rule.rhs1 == state:
		item = (rule.lhs, lspan, rspan)
		if item in outsidescores:
			out = outsidescores[item]
			if out == -1:
				n += 1
				rule = grammar.unary[state][n]
				continue
		else:
			outsidescores[item] = -1  # mark to avoid cycles
			outsidescores[item] = out = pcfgoutsidesxrec(grammar,
				insidescores, outsidescores, goal, rule.lhs, lspan, rspan)
		cost = out + rule.prob
		if cost < score:
			score = cost
		n += 1
		rule = grammar.unary[state][n]
	# could have a left sibling
	for sibsize in range(1, lspan + 1):
		n = 0
		rule = grammar.rbinary[state][n]
		while rule.rhs2 == state:
			item = (rule.lhs, lspan - sibsize, rspan)
			if item in outsidescores:
				out = outsidescores[item]
			else:
				outsidescores[item] = out = pcfgoutsidesxrec(grammar,
						insidescores, outsidescores, goal, rule.lhs,
						lspan - sibsize, rspan)
			cost = (insidescores[sibsize].get(rule.rhs1, INFINITY)
					+ out + rule.prob)
			if cost < score:
				score = cost
			n += 1
			rule = grammar.rbinary[state][n]
	# could have a right sibling
	for sibsize in range(1, rspan + 1):
		n = 0
		rule = grammar.lbinary[state][n]
		while rule.rhs1 == state:
			item = (rule.lhs, lspan, rspan - sibsize)
			if item in outsidescores:
				out = outsidescores[item]
			else:
				out = pcfgoutsidesxrec(grammar, insidescores,
					outsidescores, goal, rule.lhs, lspan, rspan - sibsize)
				outsidescores[item] = out
			cost = (insidescores[sibsize].get(rule.rhs2, INFINITY)
					+ out + rule.prob)
			if cost < score:
				score = cost
			n += 1
			rule = grammar.lbinary[state][n]
	return score


cpdef testestimates(Grammar grammar, UInt maxlen, UInt goal):
	print("getting inside")
	insidescores = inside(grammar, maxlen, {})
	for a in insidescores:
		for b in insidescores[a]:
			assert 0 <= a < grammar.nonterminals
			assert 0 <= bitlength(b) <= maxlen
			#print(a, b)
			#print("%s[%d] =" % (grammar.tolabel[a].decode('ascii'), b),
			#		exp(insidescores[a][b]))
	print(len(insidescores) * sum(map(len, insidescores.values())), '\n')
	insidescores = np.empty((grammar.nonterminals, (maxlen + 1)), dtype='d')
	insidescores[...] = np.NAN
	simpleinside(grammar, maxlen, insidescores)
	insidescores[np.isnan(insidescores)] = np.inf
	print("inside")
	for an, a in enumerate(insidescores):
		for bn, b in enumerate(a):
			if b < np.inf:
				print("%s len %d = %g" % (grammar.tolabel[an].decode('ascii'),
						bn, exp(-b)))
	#print(insidescores)
	#for a in range(maxlen):
	#	print(grammar.tolabel[goal].decode('ascii'), "len", a, "=",
	#			exp(-insidescores[goal, a]))

	print("getting outside")
	outside = np.empty((grammar.nonterminals, ) + 3 * (maxlen + 1, ), dtype='d')
	outside[...] = np.inf
	outsidelr(grammar, insidescores, maxlen, goal, outside)
	#print(outside)
	cnt = 0
	for an, a in enumerate(outside):
		for bn, b in enumerate(a):
			for cn, c in enumerate(b):
				for dn, d in enumerate(c):
					if d < np.inf:
						print("%s length %d lr %d gaps %d = %g" % (
								grammar.tolabel[an].decode('ascii'),
								bn, cn, dn, exp(-d)))
						cnt += 1
	print(cnt)
	print("done")
	return outside


def main():
	from discodop import plcfrs
	from discodop.grammar import treebankgrammar
	from discodop._grammar import Grammar
	from discodop.tree import Tree
	from discodop.treebank import NegraCorpusReader
	from discodop.treetransforms import addfanoutmarkers, binarize
	corpus = NegraCorpusReader('.', 'alpinosample.export')
	trees = list(corpus.parsed_sents().values())
	for a in trees:
		binarize(a, vertmarkov=1, horzmarkov=1)
		addfanoutmarkers(a)
	grammar = Grammar(treebankgrammar(trees, list(corpus.sents().values())))
	trees = [Tree.parse("(ROOT (A (a 0) (b 1)))", parse_leaf=int),
			Tree.parse("(ROOT (B (a 0) (c 2)) (b 1))", parse_leaf=int),
			Tree.parse("(ROOT (B (a 0) (c 2)) (b 1))", parse_leaf=int),
			Tree.parse("(ROOT (C (a 0) (c 2)) (b 1))", parse_leaf=int),
			]
	sents =[["a", "b"],
			["a", "b", "c"],
			["a", "b", "c"],
			["a", "b", "c"]]
	print("treebank:")
	for a in trees:
		print(a)
	print("\ngrammar:")
	grammar = Grammar(treebankgrammar(trees, sents))
	print(grammar, '\n')
	testestimates(grammar, 4, grammar.toid[b"ROOT"])
	outside = getestimates(grammar, 4, grammar.toid[b"ROOT"])
	sent = ["a", "b", "c"]
	print("\nwithout estimates")
	chart, start, msg = plcfrs.parse(sent, grammar, estimates=None)
	print(msg)
	plcfrs.pprint_chart(chart, sent, grammar.tolabel)
	print("\nwith estimates")
	estchart, start, msg = plcfrs.parse(sent, grammar,
			estimates=('SXlrgaps', outside))
	print(msg)
	plcfrs.pprint_chart(estchart, sent, grammar.tolabel)
	print('items avoided:')
	print(list(chart))
	print()
	print(list(estchart))
	for item in set(chart) - set(estchart):
		print(item)
	print()

	trees = [Tree.parse(a, parse_leaf=int) for a in """\
			(ROOT (A (a 0) (b 1)))
			(ROOT (A (B (A (B (a 0) (b 1))))) (c 2))
			(ROOT (A (B (A (B (a 0) (b 1))))) (c 2))
			(ROOT (A (B (A (B (a 0) (b 1))))) (c 2))
			(ROOT (A (B (A (B (a 0) (b 1))))) (c 2))
			(ROOT (A (B (A (B (a 0) (b 1))))) (c 2))
			(ROOT (C (a 0) (b 1)) (c 2))""".splitlines()]
	sents =[["a", "b"],
			["a", "b", "c"],
			["a", "b", "c"],
			["a", "b", "c"],
			["a", "b", "c"],
			["a", "b", "c"],
			["a", "b", "c"]]
	print("treebank:")
	for a in trees:
		print(a)
	print("\npcfg grammar:")
	grammar = Grammar(treebankgrammar(trees, sents))
	print(grammar, '\n')
	outside = getpcfgestimates(grammar, 4, grammar.toid[b"ROOT"], debug=True)
	sent = ["a", "b", "c"]
	print("\nwithout estimates")
	chart, start, msg = plcfrs.parse(sent, grammar, estimates=None)
	print(msg)
	plcfrs.pprint_chart(chart, sent, grammar.tolabel)
	print("\nwith estimates")
	estchart, start, msg = plcfrs.parse(sent, grammar, estimates=('SX', outside))
	print(msg)
	plcfrs.pprint_chart(estchart, sent, grammar.tolabel)
	print('items avoided:')
	for item in set(chart) - set(estchart):
		print(item)

if __name__ == '__main__':
	main()



""" Assorted functions to project items from a coarse chart to corresponding
items for a fine grammar. """
from __future__ import print_function
from collections import defaultdict
from discodop.tree import Tree
from discodop.treetransforms import mergediscnodes, unbinarize, fanout, \
		addbitsets
from discodop._grammar cimport Grammar
from discodop.containers cimport ChartItem, CFGChartItem, Edge, LCFRSEdge, \
		CFGEdge, RankedEdge, RankedCFGEdge, new_CFGChartItem, \
		CFGtoSmallChartItem, CFGtoFatChartItem, ULLong, UInt
from discodop.kbest import lazykbest
from discodop.agenda cimport Entry
import numpy as np


cpdef prunechart(chart, ChartItem goal, Grammar coarse, Grammar fine,
	int k, bint splitprune, bint markorigin, bint finecfg, bint bitpar):
	""" Produce a white list of chart items occurring in the k-best derivations
	of chart, where labels X in coarse.toid are projected to the labels
	X and X@n-m in fine.toid, for possible values of n and m.
	Modifies chart destructively.

	:param k: number of k-best derivations to consider. When k==0, the chart is
		not pruned but filtered to contain only items that contribute to a
		complete derivation.
	:param splitprune: coarse stage used a split-PCFG where discontinuous node
		appear as multiple CFG nodes. Every discontinuous node will result
		in multiple lookups into whitelist to see whether it should be
		allowed on the agenda.
	:param markorigin: in combination with splitprune, coarse labels include an
		integer to distinguish components; e.g., CFG nodes NP*0 and NP*1
		map to the discontinuous node NP_2. """
	cdef dict d
	cdef list whitelist
	cdef ChartItem Ih
	assert fine.mapping is not NULL
	if splitprune and markorigin:
		assert fine.splitmapping is not NULL
		assert fine.splitmapping[0] is not NULL
	d = <dict>defaultdict(dict)
	# construct a list of the k-best nonterminals to prune with
	if bitpar:
		kbest = bitparkbestitems(chart, coarse, finecfg)
	else:
		kbest = kbest_items(chart, goal, k, finecfg, coarse.tolabel)
	if finecfg:
		whitelist = [[{} for _ in range((<CFGChartItem>goal).end + 1)]
				for _ in range((<CFGChartItem>goal).end)]
		for left in range((<CFGChartItem>goal).end):
			for right in range(left + 1, (<CFGChartItem>goal).end + 1):
				span = left, right
				cell = whitelist[left][right]
				for label in range(1, fine.nonterminals):
					if (fine.mapping[label] == 0
							or span in kbest[fine.mapping[label]]):
						cell[label] = None
	else:
		whitelist = [None] * fine.nonterminals
		kbestspans = [{} for a in coarse.toid]
		kbestspans[0] = None
		# uses ids of labels in coarse chart
		for Ih in kbest:
			label = Ih.label
			Ih.label = 0
			kbestspans[label][Ih] = 0.0
		# now construct a list which references these coarse items:
		for label in range(fine.nonterminals):
			if splitprune and markorigin and fine.fanout[label] != 1:
				if fine.splitmapping[label] is not NULL:
					whitelist[label] = [kbestspans[fine.splitmapping[label][n]]
							for n in range(fine.fanout[label])]
			else:
				if fine.mapping[label] != 0:
					whitelist[label] = kbestspans[fine.mapping[label]]
	return whitelist, len(kbest)


cpdef kbest_items(chart, ChartItem start, int k, bint finecfg, coarsetolabel):
	""" Produce a dictionary with ChartItems as keys (value is currently
	unused) according to the k-best derivations in a chart. """
	cdef Entry entry
	cdef CFGChartItem ci
	if k == 0:
		items = filterchart(chart, start)
		for a in items:
			# use if probabilities matter
			#e = min(items[a])
			#items[a] = e.inside
			items[a] = 0.0
	else:
		_, D, agenda = lazykbest(chart, start, k, derivs=False)
		if isinstance(start, CFGChartItem):
			ci = start
			if ci.end >= (sizeof(ULLong) * 8):
				ei = CFGtoFatChartItem(ci.label, ci.start, ci.end)
			else:
				ei = CFGtoSmallChartItem(ci.label, ci.start, ci.end)
			if finecfg:
				items = [{} for _ in coarsetolabel]
				items[ci.label][ci.start, ci.end] = 0.0
			else:
				items = {ei: 0.0}
			for entry in D[(<CFGChartItem>start).start][
					(<CFGChartItem>start).end][start.label][:k]:
				traversekbestcfg(entry.key, entry.value, D, chart, items,
						(<CFGChartItem>start).end >= (sizeof(ULLong) * 8),
						finecfg, agenda)
				#agenda.difference_update(items)
		else:
			items = {start: 0.0}
			for entry in D[start][:k]:
				traversekbest(entry.key, entry.value, D, chart, items, agenda)
				#agenda.difference_update(items)
	return items


cdef traversekbest(RankedEdge ej, double rootprob, dict D, dict chart,
		dict items, set agenda):
	""" Traverse a derivation (e, j), collecting all items belonging to it, and
	noting (Viterbi) outside costs relative to its root edge (these costs are
	currently unused; only presence or absence in this list is exploited). """
	cdef LCFRSEdge e = ej.edge
	cdef RankedEdge eejj
	cdef Entry entry
	cdef double prob
	if e.left in chart:
		entry = D[e.left][ej.left]
		eejj = entry.key
		prob = entry.value
		if e.left not in items:
			items[e.left] = rootprob - prob
		if True or eejj in agenda:
			traversekbest(eejj, rootprob, D, chart, items, agenda)
	if e.right.label:
		entry = D[e.right][ej.right]
		eejj = entry.key
		prob = entry.value
		if e.right not in items:
			items[e.right] = rootprob - prob
		if True or eejj in agenda:
			traversekbest(eejj, rootprob, D, chart, items, agenda)


cdef traversekbestcfg(RankedCFGEdge ej, double rootprob, list D, list chart,
		items, bint fatitems, bint finecfg, agenda):
	""" Traverse a derivation (e, j), collecting all items belonging to it, and
	noting (Viterbi) outside costs relative to its root edge (these costs are
	currently unused; only presence or absence in this list is exploited). """
	cdef ChartItem ei
	cdef CFGEdge e = <CFGEdge>ej.edge
	cdef RankedCFGEdge eejj
	cdef Entry entry
	cdef double prob
	# TODO: efficiently track already-visited ranked edges.
	# maintaining a set only slows things down.
	if e.rule is NULL:
		return
	label = e.rule.rhs1
	start = ej.start
	end = e.mid
	if label in chart[start][end]:
		entry = D[start][end][label][ej.left]
		eejj = <RankedEdge>entry.key
		prob = entry.value
		if finecfg:
			if (start, end) not in items[label]:
				items[label][start, end] = rootprob - prob
		else:
			if fatitems:
				ei = CFGtoFatChartItem(label, start, end)
			else:
				ei = CFGtoSmallChartItem(label, start, end)
			if ei not in items:
				items[ei] = rootprob - prob
		if True or eejj in agenda:
			traversekbestcfg(eejj, rootprob, D, chart, items, fatitems,
					finecfg, agenda)
	if e.rule.rhs2:
		label = e.rule.rhs2
		start = e.mid
		end = ej.end
		entry = D[start][end][label][ej.right]
		eejj = <RankedEdge>entry.key
		prob = entry.value
		if finecfg:
			if (start, end) not in items[label]:
				items[label][start, end] = rootprob - prob
		else:
			if fatitems:
				ei = CFGtoFatChartItem(label, start, end)
			else:
				ei = CFGtoSmallChartItem(label, start, end)
			if ei not in items:
				items[ei] = rootprob - prob
		if True or eejj in agenda:
			traversekbestcfg(eejj, rootprob, D, chart, items, fatitems,
					finecfg, agenda)


cpdef filterchart(chart, ChartItem start):
	""" Remove all entries that do not contribute to a complete derivation
	headed by 'start'. """
	chart2 = {}
	if isinstance(start, CFGChartItem):
		start1 = (<CFGChartItem>start).start
		end = (<CFGChartItem>start).end
		fatitems = end >= (sizeof(ULLong) * 8)
		chart = [[b.copy() for b in a] for a in chart]
		filter_subtreecfg(start.label, start1, end, chart, chart2, fatitems)
	else:
		filter_subtree(start, chart, chart2)
	return chart2


cdef void filter_subtree(ChartItem start, dict chart, dict chart2):
	""" Recursively filter an LCFRS chart. """
	cdef LCFRSEdge edge
	cdef ChartItem item
	chart2[start] = chart[start]
	for edge in chart[start]:
		item = edge.left
		if item.label and item not in chart2:
			filter_subtree(edge.left, chart, chart2)
		item = edge.right
		if item.label and item not in chart2:
			filter_subtree(edge.right, chart, chart2)


cdef void filter_subtreecfg(label, start, end, list chart, dict chart2,
		bint fatitems):
	""" Recursively filter a PCFG chart. """
	cdef CFGEdge edge
	cdef ChartItem newitem
	if fatitems:
		newitem = CFGtoFatChartItem(label, start, end)
	else:
		newitem = CFGtoSmallChartItem(label, start, end)
	# should convert CFGEdge to LCFRSEdge, but currently not used anyway.
	chart2[newitem] = chart[start][end].pop(label)
	for edge in chart2[newitem]:
		if edge.rule is NULL:
			continue
		if edge.rule.rhs1 and edge.rule.rhs1 in chart[start][edge.mid]:
			filter_subtreecfg(edge.rule.rhs1, start, edge.mid,
					chart, chart2, fatitems)
		if edge.rule.rhs2 and edge.rule.rhs2 in chart[edge.mid][end]:
			filter_subtreecfg(edge.rule.rhs2, edge.mid, end,
					chart, chart2, fatitems)


def whitelistfromposteriors(inside, outside, CFGChartItem start,
		Grammar coarse, Grammar fine, double threshold,
		bint splitprune, bint markorigin, bint finecfg):
	""" Compute posterior probabilities and prune away cells below some
	threshold. this version is for use with parse_sparse(). """
	cdef UInt label
	cdef short lensent = start.end
	assert 0 < threshold < 1, (
			"threshold should be a cutoff for probabilities between 0 and 1.")
	sentprob = inside[0, lensent, start.label]
	posterior = (inside[:lensent, :lensent + 1]
		* outside[:lensent, :lensent + 1]) / sentprob

	finechart = [[{} for _ in range(lensent + 1)] for _ in range(lensent)]
	leftidx, rightidx, labels = (posterior[:lensent, :lensent + 1]
		> threshold).nonzero()

	kbestspans = [{} for _ in coarse.toid]
	fatitems = lensent >= (sizeof(ULLong) * 8)
	for label, left, right in zip(labels, leftidx, rightidx):
		if finecfg:
			ei = left, right
		elif fatitems:
			ei = CFGtoFatChartItem(0, left, right)
		else:
			ei = CFGtoSmallChartItem(0, left, right)
		kbestspans[label][ei] = 0.0

	if finecfg:
		whitelist = [[{} for _ in range(start.end + 1)]
				for _ in range(start.end)]
		for left in range(start.end):
			for right in range(left + 1, start.end + 1):
				span = left, right
				cell = whitelist[left][right]
				for label in range(1, fine.nonterminals):
					if (fine.mapping[label] == 0
							or span in kbestspans[fine.mapping[label]]):
						cell[label] = None
	else:
		whitelist = [None] * fine.nonterminals
		for label in range(fine.nonterminals):
			if splitprune and markorigin and fine.fanout[label] != 1:
				if fine.splitmapping[label] is not NULL:
					whitelist[label] = [kbestspans[fine.splitmapping[label][n]]
						for n in range(fine.fanout[label])]
			else:
				if fine.mapping[label] != 0:
					whitelist[label] = kbestspans[fine.mapping[label]]
	unfiltered = (outside != 0.0).sum()
	numitems = (posterior != 0.0).sum()
	numremain = (posterior > threshold).sum()
	return whitelist, sentprob, unfiltered, numitems, numremain


def whitelistfromposteriors_matrix(inside, outside, ChartItem goal,
		Grammar coarse, Grammar fine, finechart, short maxlen,
		double threshold):
	""" compute posterior probabilities and prune away cells below some
	threshold. this version produces a matrix with pruned spans having NaN as
	value. """
	cdef long label
	cdef short lensent = goal.right
	sentprob = inside[0, lensent, goal.label]
	#print >>stderr, "sentprob=%g" % sentprob
	posterior = (inside[:lensent, :lensent + 1, :]
			* outside[:lensent, :lensent + 1, :]) / sentprob
	inside[:lensent, :lensent + 1, :] = np.NAN
	inside[posterior > threshold] = np.inf
	#print >>stderr, ' ', (posterior > threshold).sum(),
	#print >>stderr, "of", (posterior != 0.0).sum(),
	#print >>stderr, "nonzero coarse items left",
	#labels, leftidx, rightidx = (posterior[:lensent, :lensent+1, :]
	#	> threshold).nonzero()
	#for left, right, label in zip(leftidx, rightidx, labels):
	#	for x in mapping[label]:
	#		finechart[left, right, x] = inside[left, right, label]
	for label in range(len(fine.toid)):
		finechart[:lensent, :lensent + 1, label] = inside[
				:lensent, :lensent + 1, fine.mapping[label]]


cpdef merged_kbest(dict chart, ChartItem start, int k, Grammar grammar):
	""" Like kbest_items, but apply the reverse of the Boyd (2007)
	transformation to the k-best derivations. """
	cdef dict newchart = <dict>defaultdict(dict)
	cdef list derivs = [Tree.parse(a, parse_leaf=int) for a, _
				in lazykbest(chart, start, k, grammar.tolabel)[0]]
	for tree in derivs:
		addbitsets(mergediscnodes(unbinarize(
				tree, childchar=":", parentchar="!")))
		for node in tree.subtrees():
			f = fanout(node)
			if f > 1:
				label = "%s_%d" % (node.label, f)
			else:
				label = node.label
			newchart[label][sum([1L << n for n in node.leaves()])] = 0.0
	return newchart


def bitparkbestitems(dict derivs, Grammar coarse, bint finecfg):
	""" Take a dictionary of CFG derivations as strings, and produce a list of
	ChartItems occurring in those derivations. """
	items = [{} for _ in coarse.tolabel] if finecfg else {}
	fatitems = None
	for deriv in derivs:
		t = Tree.parse(deriv, parse_leaf=int)
		if fatitems is None:
			fatitems = len(t.leaves()) >= (sizeof(ULLong) * 8)
		for n in t.subtrees():
			label = coarse.toid[n.label]
			leaves = n.leaves()
			start = min(leaves)
			end = max(leaves) + 1
			if finecfg:
				items[label][start, end] = 0.0
			elif fatitems:
				item = CFGtoFatChartItem(label, start, end)
				items[item] = 0.0
			else:
				item = CFGtoSmallChartItem(label, start, end)
				items[item] = 0.0
	return items


def doctf(coarse, fine, sent, tree, k, split, verbose=False):
	import plcfrs
	from disambiguation import marginalize
	from treetransforms import canonicalize, removefanoutmarkers
	from math import exp as pyexp
	sent, tags = zip(*sent)
	print(" C O A R S E ", end='')
	p, start, _ = plcfrs.parse(sent, coarse, tags=tags)
	if start:
		mpp, _, _ = marginalize("mpp", p, start, coarse, 10)
		for t in mpp:
			print(pyexp(-mpp[t]), end='')
			t = Tree.parse(t, parse_leaf=int)
			if split:
				unbinarize(t, childchar=":", parentchar="!")
				mergediscnodes(t)
			unbinarize(t)
			t = canonicalize(removefanoutmarkers(t))
			print("exact match" if t == canonicalize(tree) else "no match")
	else:
		print("no parse")
		return
		#pprint_chart(p, sent, coarse.tolabel)
	l, _ = prunechart(p, start, coarse, fine, k, split, True, False, False)
	if verbose:
		print("\nitems in 50-best of coarse chart")
		if split:
			d = merged_kbest(p, start, k, coarse)
			for label in d:
				print(label, map(bin, d[label].keys()))
		else:
			kbest = kbest_items(p, start, k, False, ())
			for a, b in kbest.items():
				print(coarse.tolabel[(<ChartItem>a).label],
						bin((<ChartItem>a).vec), b)
		print("\nwhitelist:")
		for n, x in enumerate(l):
			if isinstance(x, dict):
				print(fine.tolabel[n], map(bin, x))
			elif x:
				for m, y in enumerate(x):
					print(fine.tolabel[n], m, map(bin, y))
	print(" F I N E ", end='')
	pp, start, _ = plcfrs.parse(sent, fine, tags=tags, whitelist=l,
			splitprune=split, markorigin=True)
	if start:
		mpp, _, _ = marginalize("mpp", pp, start, fine, 10)
		for t in mpp:
			print(pyexp(-mpp[t]), end='')
			t = Tree.parse(t, parse_leaf=int)
			unbinarize(t)
			t = canonicalize(removefanoutmarkers(t))
			#print(t)
			print("exact match" if t == canonicalize(tree) else "no match", t)
	else:
		print("no parse. problem.")
		return
		#if verbose:
		#xp = set((coarse.tolabel[a.label], a.vec) for a in p.keys() if p[a])
		#xpp = set((fine.tolabel[a.label], a.vec) for a in pp.keys() if pp[a])
		#print("difference:")
		#for a in xp - xpp:
		#	if "*" not in a[0]:
		#		print(a[0], bin(a[1]))
		#print("\nfine chart:")
		#for a in xpp:
		#	if "@" not in a[0]:
		#		print(a[0], bin(a[1]))
		#pprint_chart(pp, sent, fine.tolabel)


def main():
	import re
	from treetransforms import splitdiscnodes, binarize, addfanoutmarkers
	from treebank import NegraCorpusReader
	from grammar import treebankgrammar, dopreduction, subsetgrammar
	from time import clock
	k = 50
	#corpus = NegraCorpusReader(".", "toytb.export", encoding="iso-8859-1")
	#corpus = NegraCorpusReader("../rparse", "negraproc.export",
	#	encoding="utf-8", headrules="negra.headrules", headfinal=True,
	#   headreverse=False)
	#train = 400
	#test = 40
	#testmaxlen = 999
	corpus = NegraCorpusReader('.', 'alpinosample.export')
	train = 0
	test = 3
	testmaxlen = 999
	#trees = corpus.parsed_sents().values()[:train]
	#sents = corpus.sents().values()[:train]
	trees = list(corpus.parsed_sents().values())
	sents = list(corpus.sents().values())

	dtrees = [t.copy(True) for t in trees]
	parenttrees = [t.copy(True) for t in trees]
	for t in trees:
		binarize(t, vertmarkov=1, horzmarkov=1)
		addfanoutmarkers(t)
	cftrees = [splitdiscnodes(t.copy(True), markorigin=True) for t in trees]
	for t in cftrees:
		binarize(t, horzmarkov=2, tailmarker='', leftmostunary=True,
			childchar=":")  # NB leftmostunary is important
		addfanoutmarkers(t)
	for t in parenttrees:
		binarize(t, vertmarkov=3, horzmarkov=1)
		addfanoutmarkers(t)
	for t in dtrees:
		binarize(t, vertmarkov=1, horzmarkov=1)
		addfanoutmarkers(t)
	normallcfrs = treebankgrammar(trees, sents)
	normal = Grammar(normallcfrs)
	parent = Grammar(treebankgrammar(parenttrees, sents))
	splitg = Grammar(treebankgrammar(cftrees, sents))
	for t, s in zip(cftrees, sents):
		for (r, yf), w in treebankgrammar([t], [s]):
			assert len(yf) == 1
	fine999x = dopreduction(trees, sents)[0]
	fine999 = Grammar(fine999x)
	fine1 = Grammar(dopreduction(dtrees, sents)[0])
	trees = list(corpus.parsed_sents().values())[train:train + test]
	sents = list(corpus.tagged_sents().values())[train:train + test]
	if subsetgrammar(normallcfrs, fine999x):
		print("DOP grammar is a superset")
	else:
		print("DOP grammar is NOT a superset!")
	for msg, coarse, fine, split, enabled in zip(
		"normal parentannot cf-split".split(),
		(normal, parent, splitg),
		(fine999, fine1, fine1),
		(False, False, True),
		(True, False, True)):
		if not enabled:
			continue
		print("coarse grammar:", msg)
		fine.getmapping(coarse, re.compile(b'@[-0-9]+$'), None, split, True)
		begin = clock()
		for n, (sent, tree) in enumerate(zip(sents, trees)):
			if len(sent) > testmaxlen:
				continue
			print(n, end='')
			doctf(coarse, fine, sent, tree, k, split, verbose=False)
		print("time elapsed", clock() - begin, "s")

if __name__ == '__main__':
	main()
